#!/usr/bin/env python3
#
# Usage
# ~~~~~
# aiko_registrar
# ./share.py sc_test  # ServicesCache test
#
# ./share.py ec_test  # EventualConsistency test
# ./share.py ec_test PRODUCER_PID [EC_PRODUCER_SID]
#
# NAMESPACE=aiko
# HOST=localhost
# PRODUCER_PID=`ps ax | grep python | grep share.py | xargs | cut -d" " -f1`;  \
#     TOPIC_PATH=$NAMESPACE/$HOST/$PRODUCER_PID/1
#
# mosquitto_sub -t '#' -v
#
# mosquitto_pub -t $TOPIC_PATH/control -m "(invalid)"
# mosquitto_pub -t $TOPIC_PATH/control -m "(share)"
# mosquitto_pub -t $TOPIC_PATH/control -m "(share topic)"
# mosquitto_pub -t $TOPIC_PATH/control -m "(share topic 0)"
# mosquitto_pub -t $TOPIC_PATH/control -m "(share topic a *)"
# mosquitto_pub -t $TOPIC_PATH/control -m "(share topic 0 *)"
# mosquitto_pub -t $TOPIC_PATH/control -m "(share topic 0 lifecycle)"
# mosquitto_pub -t $TOPIC_PATH/control -m "(share topic 0 services)"
# mosquitto_pub -t $TOPIC_PATH/control -m "(share topic 0 x)"
# mosquitto_pub -t $TOPIC_PATH/control -m "(share topic 0 (lifecycle))"
# mosquitto_pub -t $TOPIC_PATH/control -m "(share topic 0 (lifecycle x))"
# mosquitto_pub -t $TOPIC_PATH/control -m "(add count 0)"
# mosquitto_pub -t $TOPIC_PATH/control -m "(update count 1)"
# mosquitto_pub -t $TOPIC_PATH/control -m "(remove count)"
# mosquitto_pub -t $TOPIC_PATH/control -m "(update lifecycle ready)"
# mosquitto_pub -t $TOPIC_PATH/control -m "(add services.test 0)"
# mosquitto_pub -t $TOPIC_PATH/control -m "(remove services)"
#
# Environment variables
# ~~~~~~~~~~~~~~~~~~~~~
# Fine-grained logging control ...
# - Show some low-level EventualConsistency and ServicesCache operations
#   - AIKO_LOG_LEVEL_SHARE: ERROR, WARNING, INFO, DEBUG
#
# To Do
# ~~~~~
# - BUG?: ECProducer remote expired leases
# - BUG?: ECConsumer remote expired leases
#
# - BUG: ServicesCache (and also EC) should handle degradation of
#   network (ConnectionState changes) and/or lease expiring.
#
# - ECProducerCore class provides absolute minimum implementation
#   - Responds to all "(share ...)" requests with "(item_count 0)"
#     Ignores all other requests
#   - ECProducer extends ECProducerCore
#
# - For multiple Actors in same process, when each Actor wants to have a
#   its own unique ECProducer shared amongst different aspects of the Actor,
#   e.g LifeCycleClient and some other functionality, then the ECProducer
#   constructor optionally includes the Service "name" to disambigute
#
# - Provide unit tests !
# - Registrar should migrate use of ServicesCache to ECProducer
# - ECProducer and ECConsumer should handle Registrar not available
# - ECProducer and ECConsumer should handle Registrar stop and restart
# - Allow ECConsumer to change share filter, with or without existing lease
# - Fix ECConsumer to handle _ec_remove_item()/_ec_update_item() exceptions
# - Improve EC share/cache to work recursively beyond just two levels
# - When subscribed to a component of a dictionary, removal of the dictionary
#   does not send a remove message for the component

import click
import os

from aiko_services.main import *
from aiko_services.main.utilities import *

__all__ = [
    "ECConsumer", "PROTOCOL_EC_CONSUMER",
    "ECProducer", "PROTOCOL_EC_PRODUCER",
    "services_cache_create_singleton", "services_cache_delete"
]

_VERSION = 0

SERVICE_TYPE_EC_CONSUMER = "ec_consumer_test"
PROTOCOL_EC_CONSUMER =  \
    f"{SERVICE_PROTOCOL_AIKO}/{SERVICE_TYPE_EC_CONSUMER}:{_VERSION}"

SERVICE_TYPE_EC_PRODUCER = "ec_producer_test"
PROTOCOL_EC_PRODUCER =  \
    f"{SERVICE_PROTOCOL_AIKO}/{SERVICE_TYPE_EC_PRODUCER}:{_VERSION}"

_LEASE_TIME = 300  # seconds

_AIKO_LOG_LEVEL_SHARE = os.environ.get("AIKO_LOG_LEVEL_SHARE", "INFO")
_LOGGER = aiko.logger(__name__, log_level=_AIKO_LOG_LEVEL_SHARE)

# --------------------------------------------------------------------------- #

def _ec_modify_item(items, item_path, modify_operation, create_path=False):
    if not isinstance(items, dict):
        item_type = type(items).__name__
        raise ValueError(f'"items" must be a dictionary, not {item_type}')
    if not isinstance(item_path, list):
        item_path_type = type(item_path).__name__
        raise ValueError(f'"item_path" must be a list, not {item_path_type}')
    if len(item_path) == 0:
        raise ValueError('"item_path" must be non-empty')

    item_key, *item_path_tail = item_path
    if len(item_path_tail) == 0:
        modify_operation(items, item_key)
    else:
        if item_key in items:
            nested_items = items[item_key]
            _ec_modify_item(nested_items, item_path_tail, modify_operation)
        elif create_path:
            nested_items = {}
            items[item_key] = nested_items
            _ec_modify_item(nested_items, item_path_tail, modify_operation)

def _ec_parse_item_path(name):
   item_path = name.split(".")
   if len(item_path) > 2:
       raise ValueError(f'EC "share" dictionary depth maximum is 2: {name}')
   return item_path

def _ec_remove_item(share, item_path):
    def remove_item(items, item_key):
        if item_key in items:
            del items[item_key]
    _ec_modify_item(share, item_path, remove_item)

def _ec_update_item(share, item_path, item_value):
    def update_item(items, item_key):
        items[item_key] = item_value
    _ec_modify_item(share, item_path, update_item, create_path=True)

def _flatten_dictionary(dictionary):
    result = []
    for item_name, item in dictionary.items():
        if isinstance(item, dict):
            for subitem_name, subitem in item.items():
                name = f"{item_name}.{subitem_name}"
                result.append((name, subitem))
        else:
            result.append((item_name, item))
    return result

# --------------------------------------------------------------------------- #

class ECLease(Lease):
    def __init__(
        self, lease_time, topic, filter=None, lease_expired_handler=None):

        super().__init__(
            lease_time, topic, lease_expired_handler=lease_expired_handler)
        self.filter = filter

class ECProducer:
    def __init__(self, service, share, topic_in=None, topic_out=None):
        self.share = share
        self.topic_in = topic_in if topic_in else service.topic_control
        self.topic_out = topic_out if topic_out else service.topic_state
        self.handlers = set()
        self.leases = {}
        service.add_message_handler(self._producer_handler, self.topic_in)
        service.add_tags(["ec=true"])

    def add_handler(self, handler):
        for item_name, item_value in _flatten_dictionary(self.share):
            handler("add", item_name, item_value)
        self.handlers.add(handler)

    def get(self, item_name):
        success = True
        item_path = _ec_parse_item_path(item_name)
        item = self.share
        for key in item_path:
            if isinstance(item, dict) and key in item:
                item = item.get(key)
            else:
                success = False
        if not success:
            item = None
        return item

    def update(self, item_name, item_value):
        item_path = _ec_parse_item_path(item_name)
        success = False
        try:
            _ec_update_item(self.share, item_path, item_value)
            success = True
        except ValueError as value_error:
            diagnostic = f'update {item_name}: {value_error}'
            _LOGGER.error(f"update(): {diagnostic}")
        if success:
            self._update_consumers("update", item_name, item_value)

    def remove(self, item_name):
        item_path = _ec_parse_item_path(item_name)
        success = False
        try:
            _ec_remove_item(self.share, item_path)
            success = True
        except ValueError as value_error:
            diagnostic = f'remove {item_name}: {value_error}'
            _LOGGER.error(f"remove(): {diagnostic}")
        if success:
            self._update_consumers("remove", item_name, None)

    def remove_handler(self, handler):
        if handler in self.handlers:
            self.handlers.remove(handler)

    def _dictionary_to_commands(self, command, dictionary):
        payloads = []
        for name, subitem in _flatten_dictionary(dictionary):
            payloads.append(generate(command, [name, subitem]))
        return payloads

    def _ec_parse_share(self, command, parameters):
        response_topic = None
        lease_time = None
        filter = []

        if command == "share" and len(parameters) == 3:
            try:
                lease_time = int(parameters[1])
            except Exception:
                pass
            if isinstance(lease_time, int):
                filter = parameters[2]
                if filter != "*" and not isinstance(filter, list):
                    filter = [filter]
                response_topic = parameters[0]  # Deliberately do this last

        return response_topic, lease_time, filter

    def _filter_compare(self, filter, item_name):
        match = False
        if filter == "*":
            match = True
        else:
            for filter_item in filter:
                if item_name == filter_item or \
                   item_name.startswith(f"{filter_item}."):
                    match = True
        return match

    def _lease_expired_handler(self, topic):
        if topic in self.leases:
            del self.leases[topic]

    def _producer_handler(self, aiko, topic, payload_in):
        command, parameters = parse(payload_in)
        payload_out = payload_in

        if (command == "add" or command == "update") and len(parameters) == 2:
            item_name = parameters[0]
            item_value = parameters[1]
            item_path = _ec_parse_item_path(item_name)

            success = False
            try:
                _ec_update_item(self.share, item_path, item_value)
                success = True
            except ValueError as value_error:
                diagnostic = f'command "{command}", {parameters}: {value_error}'
                _LOGGER.error(f"_producer_hander(): {diagnostic}")
            if success:
                aiko.message.publish(self.topic_out, payload_out)
                self._update_consumers(command, item_name, item_value)

        if command == "remove" and len(parameters) == 1:
            item_name = parameters[0]
            item_path = _ec_parse_item_path(item_name)

            success = False
            try:
                _ec_remove_item(self.share, item_path)
                success = True
            except ValueError as value_error:
                diagnostic = f'command "{command}", {parameters}: {value_error}'
                _LOGGER.error(f"_producer_hander(): {diagnostic}")
            if success:
                aiko.message.publish(self.topic_out, payload_out)
                self._update_consumers(command, item_name, None)

        if command == "share":
            response_topic, lease_time, filter = self._ec_parse_share(
                command, parameters)
            if response_topic:
                if lease_time == 0:
                    if response_topic in self.leases:
                        self.leases[response_topic].terminate()
                        del self.leases[response_topic]
                    else:
                        self._synchronize(response_topic, filter)
                elif lease_time > 0:
                    if response_topic in self.leases:
                        self.leases[response_topic].extend(lease_time)
                    else:
                        lease = ECLease(
                            lease_time,
                            response_topic,
                            filter=filter,
                            lease_expired_handler=self._lease_expired_handler)
                        self.leases[response_topic] = lease
                        self._synchronize(response_topic, filter)

    def _filter_dictionary(self, dictionary, filter, path):
        share = {}
        for item_name, item in dictionary.items():
            item_path = path + [str(item_name)]
            if isinstance(item, dict):
                filtered_item = self._filter_dictionary(item, filter, item_path)
                if filtered_item != {}:
                    share[item_name] = filtered_item
            else:
                item_path_str = ".".join(item_path)
                if self._filter_compare(filter, item_path_str):
                    share[item_name] = item
        return share

    def _filter_share(self, filter):
        return self._filter_dictionary(self.share, filter, [])

    def _synchronize(self, response_topic, filter):
        filtered_share = self._filter_share(filter)
        commands = self._dictionary_to_commands("add", filtered_share)

        command_count = len(commands)
        payload_out = f"(item_count {command_count})"
        aiko.message.publish(response_topic, payload_out)
        for payload_out in commands:
            aiko.message.publish(response_topic, payload_out)

        payload_out = f"(sync {response_topic})"
        aiko.message.publish(self.topic_out, payload_out)

    def _update_consumers(self, command, item_name, item_value):
        for handler in self.handlers:
            handler(command, item_name, item_value)
        if command == "remove":
            payload_out = f"({command} {item_name})"
        else:
            # TODO: Use "generate()" on item_value
            payload_out = f"({command} {item_name} {item_value})"
        for lease in self.leases.values():
            if self._filter_compare(lease.filter, item_name):
                response_topic = lease.lease_uuid
                aiko.message.publish(response_topic, payload_out)

# --------------------------------------------------------------------------- #
# Note: For non-Service use, can substitute "aiko.process" for "service"

class ECConsumer:
    def __init__(self,
        service, ec_consumer_id, cache, ec_producer_topic_control, filter="*"):

        self.service = service
        self.ec_consumer_id = ec_consumer_id
        self.cache = cache
        self.ec_producer_topic_control = ec_producer_topic_control
        self.filter = filter

        self.cache_state = "empty"
        self.handlers = set()
        self.item_count = 0
        self.items_received = 0
        self.lease = None

        self.topic_share_in = \
            f"{self.service.topic_path}/{self.ec_producer_topic_control}/{self.ec_consumer_id}/in"
        self.service.add_message_handler(self._consumer_handler, self.topic_share_in)
        aiko.connection.add_handler(self._connection_state_handler)

    def add_handler(self, handler):
        for item_name, item_value in _flatten_dictionary(self.cache):
            handler(self.ec_consumer_id, "add", item_name, item_value)
        self.handlers.add(handler)

    def _consumer_handler(self, aiko, topic, payload_in):
        command, parameters = parse(payload_in)

        if command == "item_count" and len(parameters) == 1:
            self.item_count = parse_int(parameters[0])
            self.items_received = 0

        elif command == "add" and len(parameters) == 2:
            item_name = parameters[0]
            item_path = _ec_parse_item_path(item_name)
            item_value = parameters[1]
# TODO: Catch ValueError and provide better console log (see ECProducer)
            _ec_update_item(self.cache, item_path, item_value)
            self.items_received += 1
            if self.items_received == self.item_count:
                self.cache_state = "ready"
            self._update_handlers(command, item_name, item_value)

        elif command == "remove" and len(parameters) == 1:
            item_name = parameters[0]
            item_path = _ec_parse_item_path(item_name)
# TODO: Catch ValueError and provide better console log (see ECProducer)
            _ec_remove_item(self.cache, item_path)
            self._update_handlers(command, item_name, None)

        elif command == "update" and len(parameters) == 2:
            item_name = parameters[0]
            item_path = _ec_parse_item_path(item_name)
            item_value = parameters[1]
# TODO: Catch ValueError and provide better console log (see ECProducer)
            _ec_update_item(self.cache, item_path, item_value)
            self._update_handlers(command, item_name, item_value)
        elif command == "sync":
            #pass  # ECConsumer doesn't use the share "(sync ...)" command
            self._update_handlers(command, None, None)
        else:
            diagnostic = f"Unknown command: {command}, {parameters}"
            _LOGGER.debug(f"_consumer_handler(): {diagnostic}")

     #   for item_name, item in self.cache.items():
     #       _LOGGER.debug(f"ECConsumer cache {item_name}: {item}")
     #   _LOGGER.debug("----------------------------")

    def _connection_state_handler(self, connection, connection_state):
        if connection.is_connected(ConnectionState.REGISTRAR):
            if not self.lease:
                self.lease = Lease(
                    _LEASE_TIME, None, automatic_extend=True,
                    lease_extend_handler=self._share_request)
                self._share_request()

    def remove_handler(self, handler):
        if handler in self.handlers:
            self.handlers.remove(handler)

    def _share_request(self, lease_time=_LEASE_TIME, lease_uuid=None):
        aiko.message.publish(
            self.ec_producer_topic_control,
            f"(share {self.topic_share_in} {lease_time} {self.filter})"
        )

    def _update_handlers(self, command, item_name, item_value):
        for handler in self.handlers:
            handler(self.ec_consumer_id, command, item_name, item_value)

    def terminate(self):
        self.service.remove_message_handler(
            self._consumer_handler, self.topic_share_in)
        aiko.connection.remove_handler(self._connection_state_handler)
        self.cache = {}
        self.cache_state = "empty"

        if self.lease:
            self.lease.terminate()
            self.lease = None
            self._share_request(lease_time=0)  # cancel share request

# --------------------------------------------------------------------------- #
# Note: For use by non-Service, can substitute "aiko.process" for "service"
#
# Service cache states
# ~~~~~~~~~~~~~~~~~~~~
# - empty:   No history or running Services and waiting for Service Registrar
# - history: Waiting for Services history to be shared
# - share:   Waiting for Services running to be shared
# - loaded:  Services populated and waiting for Service Registrar "sync" message
# - ready:   Ready for use and continuously updating
#
# To Do
# ~~~~~
# - Reimplement ServicesCache using ECProducer / ECConsumer
# - aiko_services.registrar should have identical code to the cache code !
# - services_cache_delete() should be a class method delete()

from collections import deque
from threading import Thread
import time

_HISTORY_RING_BUFFER_SIZE = 4096

class ServicesCache():
    def __init__(self, service, event_loop_start=False, history_limit=0):
        self._service = service
        self._event_loop_start = event_loop_start
        self._event_loop_owner = False
        self._history_limit = history_limit

        self._cache_reset()
        self._handlers = set()
        self._history = deque(maxlen=_HISTORY_RING_BUFFER_SIZE)
        self._registrar_topic_share = f"{service.topic_path}/registrar_share"
        aiko.connection.add_handler(self._connection_state_handler)

    def _cache_reset(self):
        self._begin_registration = False
        self._item_count = None
        self._registrar_service = None
        self._registrar_topic_in = None
        self._registrar_topic_out = None
        self._services = Services()
        self._state = "empty"

    def add_handler(self, service_change_handler, service_filter):
        if self._state in ["loaded", "ready"]:
        # TODO: Provide filtered Services to "service_change_handler"
            service_change_handler("sync", None)
        self._handlers.add((service_change_handler, service_filter))

    def remove_handler(self, service_change_handler, service_filter):
        if (service_change_handler, service_filter) in self._handlers:
            self._handlers.remove((service_change_handler, service_filter))

    def _connection_state_handler(self, connection, connection_state):
        if connection.is_connected(ConnectionState.REGISTRAR):
            if not self._begin_registration:
                self._begin_registration = True
                self._registrar_topic_in = f"{aiko.registrar['topic_path']}/in"
                self._registrar_topic_out = f"{aiko.registrar['topic_path']}/out"
                self._service.add_message_handler(
                    self.registrar_out_handler, self._registrar_topic_out
                )
                self._service.add_message_handler(
                    self.registrar_share_handler, self._registrar_topic_share
                )
                if self._history_limit > 0:
                    self._publish_registrar_history()
                    self._state = "history"
                else:
                    self._publish_registrar_share()
                    self._state = "share"
        else:
            if self._registrar_topic_out:
                self._service.remove_message_handler(
                    self.registrar_out_handler, self._registrar_topic_out
                )
                self._service.remove_message_handler(
                    self.registrar_share_handler, self._registrar_topic_share
                )
                if self._registrar_service:
                    self._history.appendleft(self._registrar_service)
                self._cache_reset()

    def _publish_registrar_history(self):
        aiko.message.publish(
            self._registrar_topic_in,
            f"(history {self._registrar_topic_share} {self._history_limit})"
        )

    def _publish_registrar_share(self):
        aiko.message.publish(
            self._registrar_topic_in,
            f"(share {self._registrar_topic_share} * * * * *)"
        )

    def _update_handlers(self, command, service_details=None):
        topic_path = service_details[0] if service_details else None
        for handler, filter in self._handlers:
            if topic_path:
                services = self._services.filter_services(filter)
                service = services.get_service(topic_path)
            else:
                service = True
            if service:
                handler(command, service_details)

    def get_history(self):
        return self._history

    def get_services(self):
        return self._services

    def get_state(self):
        return self._state

    def registrar_share_handler(self, aiko, topic_path, payload_in):
        command, parameters = parse(payload_in)
        if command == "item_count" and len(parameters) == 1:
            self._item_count = int(parameters[0])
        elif command == "add" and len(parameters) >= 6:
            if self._item_count is None:
                _LOGGER.error("Service cache: registrar_share_handler(): "
                    "(item_count N) was incorrect")
                return
            self._item_count -= 1
            service_details = parameters
            if self._state == "history":
                self._history.append(service_details)
            elif self._state == "share":
                service_topic_path = service_details[0]
                self._services.add_service(service_topic_path, service_details)
                if service_topic_path == aiko.registrar["topic_path"]:
                    self._registrar_service = service_details
        else:
            _LOGGER.debug(f"Service cache: registrar_share_handler(): Unhandled message topic: {topic_path}, payload: {payload_in}")

        if self._item_count == 0:
            self._item_count = None
            if self._state == "history":
                self._publish_registrar_share()
                self._state = "share"
            elif self._state == "share":
                self._state = "loaded"
                self._update_handlers("sync")
                for service_details in self._services:
                    self._update_handlers("add", service_details)

    def registrar_out_handler(self, aiko, topic, payload_in):
        command, parameters = parse(payload_in)
        if command == "sync" and len(parameters) == 1:
            sync_topic = parameters[0]
            if sync_topic == self._registrar_topic_share and self._state == "loaded":
                self._state = "ready"
        elif command == "add" and len(parameters) == 6:
            service_details = parameters
            self._services.add_service(service_details[0], service_details)
            self._update_handlers(command, service_details)
        elif command == "remove":
            topic_path = parameters[0]
            service_details = self._services.get_service(topic_path)
            if service_details:
                self._update_handlers(command, service_details)
                self._services.remove_service(topic_path)
                self._history.appendleft(service_details)
        else:
            _LOGGER.debug(f"Service cache: registrar_out_handler(): Unknown command: topic: {topic}, payload: {payload_in}")

    def run(self):
        if self._event_loop_start and not event.event_loop_running:
            self._event_loop_owner = True
            diagnostic = None
            try:
                aiko.process.run()
            except SystemError:
                diagnostic = "Error: MQTT Connection Reset: Incorrect configuration or security credentials ?"
            if diagnostic:
                _LOGGER.error(diagnostic)
                raise ValueError(diagnostic)  # TODO: Provide stack trackback ?

    def terminate(self):
        if self._event_loop_owner:
            aiko.process.terminate()

    def wait_ready(self):
        while self._state != "ready":
            time.sleep(0.1)

services_cache = None

def services_cache_create_singleton(
    service, event_loop_start=False, history_limit=0):

    global services_cache

    if not services_cache:
        services_cache = ServicesCache(service, event_loop_start, history_limit)
        Thread(target=services_cache.run).start()
    return services_cache

def services_cache_delete():
    global services_cache

    if services_cache:
        services_cache.terminate()
        services_cache = None

# --------------------------------------------------------------------------- #

class ECProducerTest(Service):
    def __init__(self, context):
        context.get_implementation("Service").__init__(self, context)
        _LOGGER.info(f"ECProducer: topic path: {self.topic_path}")

        self.share = {
            "lifecycle": "ready",
            "log_level": get_log_level_name(_LOGGER),
            "source_file": f"v{_VERSION}⇒ {__file__}",
            "items": {
                "key_1": ["item_1a", "item_1b"],
                "key_2": ["item_2a", "item_2b"]
            }
        }
        self.ec_producer = ECProducer(self, self.share)
        self.ec_producer.add_handler(self._ec_producer_change_handler)

    def _ec_producer_change_handler(self, command, item_name, item_value):
        _LOGGER.info(f"ECProducer: {command} {item_name} {item_value}")
        if item_name == "log_level":
            _LOGGER.setLevel(str(item_value).upper())

# --------------------------------------------------------------------------- #

class ECConsumerTest(Service):
    def __init__(self, context, ec_producer_pid, ec_producer_sid, filter):
        context.get_implementation("Service").__init__(self, context)
        _LOGGER.info(f"ECConsumer: topic path: {self.topic_path}")

        self.share_producer = {
            "lifecycle": "ready",
            "log_level": get_log_level_name(_LOGGER),
            "source_file": f"v{_VERSION}⇒ {__file__}",
            "ec_producer_pid": ec_producer_pid,
            "ec_producer_sid": ec_producer_sid
        }
        self.ec_producer = ECProducer(self, self.share_producer)

        self.share_consumer = {}
        ec_producer_topic_control =  \
            f"{get_namespace()}/{get_hostname()}/{ec_producer_pid}/{ec_producer_sid}/control"
        self.ec_consumer = ECConsumer(
            self, 0, self.share_consumer, ec_producer_topic_control, filter)
        self.ec_consumer.add_handler(self._ec_consumer_change_handler)

    def _ec_consumer_change_handler(
        self, client_id, command, item_name, item_value):
        _LOGGER.info(
            f"ECConsumer: {client_id}: {command} {item_name} {item_value}")

# --------------------------------------------------------------------------- #

@click.group()

def main():
    pass

@main.command("sc_test",
    help=("Test Registrar Services Cache"))

def sc_test():
    services_cache = services_cache_create_singleton(
        aiko.process, True, history_limit=4)

    _LOGGER.info("ServicesCache: Wait ready")
    services_cache.wait_ready()

    _LOGGER.info("ServicesCache: Services running")
    services = services_cache.get_services()
    for service_details in services:
        _LOGGER.info(f"{service_details}")

    _LOGGER.info("ServicesCache: Service history")
    history = services_cache.get_history()
    for service_details in history:
        _LOGGER.info(f"{service_details}")

    aiko.process.terminate()

@main.command("ec_test",
    help=("Test Eventual Consistency Producer and Consumer"))
@click.argument("ec_producer_pid", nargs=1, required=False)
@click.argument("ec_producer_sid", nargs=1, required=False, default="1")
@click.argument("filter", nargs=1, default="*", required=False)

def ec_test(ec_producer_pid, ec_producer_sid, filter):
    tags = ["ec=true"]  # TODO: Add ECProducer tag before add to Registrar

    if ec_producer_pid:
        init_args = service_args(
            SERVICE_TYPE_EC_CONSUMER, None, None, PROTOCOL_EC_CONSUMER, tags)
        init_args["ec_producer_pid"] = ec_producer_pid
        init_args["ec_producer_sid"] = ec_producer_sid
        init_args["filter"] = filter
        ec_consumer_test = compose_instance(ECConsumerTest, init_args)
    else:
        init_args = service_args(
            SERVICE_TYPE_EC_PRODUCER, None, None, PROTOCOL_EC_PRODUCER, tags)
        ec_producer_test = compose_instance(ECProducerTest, init_args)

    aiko.process.run(True)

if __name__ == "__main__":
    main()

# --------------------------------------------------------------------------- #



# To Do
# ~~~~~
# - None, yet !

from abc import abstractmethod

import aiko_services as aiko

__all__ = ["DataScheme"]

# --------------------------------------------------------------------------- #

class DataScheme:
    LOOKUP = {}  # key: name, value: class

    def __init__(self, pipeline_element):
        self.pipeline_element = pipeline_element
        self.share = pipeline_element.share

    @classmethod
    def add_data_scheme(cls, name, data_scheme_class):
        if name in DataScheme.LOOKUP:
            raise RuntimeError(
                f'DataScheme.add_data_scheme(): scheme "{name}" already exists')
        DataScheme.LOOKUP[name] = data_scheme_class

    @classmethod
    def contains_all(cls, source: str, match: chr):
        return False not in [character in source for character in match]

    @abstractmethod
    def create_sources(self, stream, data_sources,
        frame_generator=None, use_create_frame=True):
        return aiko.StreamEvent.OKAY, {}

    def create_targets(self, stream, data_targets):
        return aiko.StreamEvent.OKAY, {}

    def destroy_sources(self, stream):
        pass

    def destroy_targets(self, stream):
        pass

    @classmethod
    def parse_data_url_path(cls, data_url):  # data_source or data_target
        tokens = data_url.split("://")  # URL "scheme://path" or "path"
        path = tokens[0] if len(tokens) == 1 else tokens[1]
        return path

    @classmethod
    def parse_data_url_scheme(cls, data_url):  # data_source or data_target
        tokens = data_url.split("://")  # URL "scheme://path"
        scheme = "file" if len(tokens) == 1 else tokens[0]
        return scheme.lower()

# --------------------------------------------------------------------------- #



from typing import List

from ..core import paginate
from .base_models import ImageConnection
from .base_models import ImagePresigned as ImagePresignedType
from .gql_client import HLClient

__all__ = [
    "get_presigned_url",
    "get_presigned_urls",
]


def get_presigned_url(
    client: HLClient,
    id: int,
):
    """Return a single ImagePresignedType BaseModel
    for the given file id
    """
    result = client.image(
        return_type=ImagePresignedType,
        id=id,
    )
    return result


def get_presigned_urls(
    client,
    ids: List[int],
):
    """Return a generator of ImagePresignedType GQLBaseModels
    for the given list of file ids
    """
    return paginate(
        client.imageConnection,
        ImageConnection,
        id=ids,
    )



# Usage
# ~~~~~
# - gst-launch-1.0 videotestsrc ! video/x-raw,width=852,height=480,framerate=\(fraction\)24/1 ! x264enc pass=pass1 threads=0 bitrate=900 tune=zerolatency ! flvmux name=mux ! rtmpsink location="rtmp://highlighter.antmedia.io/WebRTCAppEE/759246100812675440813771"
#
# To Do
# ~~~~~
# - Replace "Gst.parse_launch()" with hand-built pipeline !

import sys
import time
from threading import Thread

if sys.version_info >= (3,0):
  from queue import Queue, Empty
else:
  from Queue import Queue, Empty

from aiko_services.elements.gstreamer import *

__all__ = ["VideoStreamWriter"]

# -----------------------------------------------------------------------------

data = None

class VideoStreamWriter:
  def __init__(self, hostname, port, width, height, framerate, rtmp_url=None):
    self.Gst = gst_initialise()
    self.queue = Queue(maxsize=30)

    if rtmp_url:
      gst_launch_command = 'appsrc name=source ! videoconvert ! x264enc bitrate=1000 me=4 subme=10 ref=2 tune=zerolatency ! video/x-h264 ! h264parse ! video/x-h264 ! queue ! flvmux name=muxer streamable=true ! rtmpsink location="{}" sync=false'.format(rtmp_url)
    else:
      gst_launch_command = "appsrc name=source ! videoconvert ! {} {} ! rtph264pay config-interval=5 pt=96 ! udpsink host={} port={}".format(utilities.get_h264_encoder(), utilities.get_h264_encoder_options(), hostname, port)

    pipeline = self.Gst.parse_launch(gst_launch_command)
    appsrc = pipeline.get_child_by_name("source")
    src_caps = "video/x-raw,format={},width={},height={},framerate={}".format(utilities.get_format(), width, height, framerate)
    appsrc.props.caps = self.Gst.caps_from_string(src_caps)

    appsrc.props.stream_type  = 0  # GST_APP_STREAM_TYPE_STREAM
    appsrc.props.is_live      = True
    appsrc.props.do_timestamp = True
    appsrc.props.format       = self.Gst.Format.TIME

    pipeline.set_state(self.Gst.State.PLAYING)

    def _run_consume_queue():
      global data

      while True:
        frame = self.queue.get()
        if "type" in frame and frame["type"] == "EOS":
          appsrc.emit('end-of-stream')
        else:
#         print("VSW frame: ", frame["id"], self.queue.qsize())
          data = frame["image"].tobytes()
          buffer = self.Gst.Buffer.new_allocate(None, len(data), None)
          buffer.fill(0, data)
          appsrc.emit('push-buffer', buffer)

    self._t1 = Thread(target=_run_consume_queue, args=())
    self._t1.daemon = True
    self._t1.start()

    def _run_constant_framerate():
    # global data  # TODO: flake8: F824 `global data` is unused

      while True:
        time.sleep(0.04)
        if data:
          buffer = self.Gst.Buffer.new_allocate(None, len(data), None)
          buffer.fill(0, data)
          appsrc.emit('push-buffer', buffer)

#   self._t2 = Thread(target=_run_constant_framerate, args=())
#   self._t2.daemon = True
#   self._t2.start()

  def queue_size(self):
    return self.queue.qsize()

  def write_frame(self, frame):
    self.queue.put(frame)



import logging
import os

__all__ = ["get_default_logger", "LOG"]


def get_default_logger(name):
    # https://stackoverflow.com/questions/43109355/logging-setlevel-is-being-ignored
    logging.debug(f"Setting up logging for logger={name}")
    logger = logging.getLogger(name)
    logger.setLevel(level=os.environ.get("LOG_LEVEL", "INFO"))
    return logger


class ColourStr:
    HEADER = "\033[95m"

    BLUE = "\033[94m"
    CYAN = "\033[96m"
    GREEN = "\033[92m"
    ORANGE = "\033[93m"
    RED = "\033[91m"

    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"

    RESET = "\033[0m"

    @staticmethod
    def blue(s):
        return ColourStr.BLUE + s + ColourStr.RESET

    @staticmethod
    def cyan(s):
        return ColourStr.CYAN + s + ColourStr.RESET

    @staticmethod
    def green(s):
        return ColourStr.GREEN + s + ColourStr.RESET

    @staticmethod
    def red(s):
        return ColourStr.RED + s + ColourStr.RESET

    @staticmethod
    def bold(s):
        return ColourStr.BOLD + s + ColourStr.RESET

    @staticmethod
    def underline(s):
        return ColourStr.UNDERLINE + s + ColourStr.RESET


LOG = get_default_logger("hl_client")



# Environment variables
# ~~~~~~~~~~~~~~~~~~~~~
# Fine-grained logging control ...
# - Show all low-level Lease operations
#   - AIKO_LOG_LEVEL_LEASE: ERROR, WARNING, INFO, DEBUG
#
# Example
# ~~~~~~~
# from aiko_services.main import *
#
# def lease_expired(lease_uuid):
#     if lease_uuid in leases:
#         del leases[lease_uuid]
#
# leases = {}
# lease_time = 10  # seconds
# lease_uuid = "1"
# lease = Lease(lease_time, lease_uuid,
#     lease_expired_handler=lease_expired, automatic_extend=True)
# leases[lease_uuid] = lease
# event.loop()
# lease.extend()
#
# To Do
# ~~~~~
# - Change "lease_time: seconds" to "lease: {start: UTC, duration: seconds}" ?
#
# - Lease request rejected ...
#   - Response (lease_rejected maximum_time 60
#   - Response (lease_rejected no_resource memory)

import os

from aiko_services.main import *
from aiko_services.main.utilities import *

__all__ = ["Lease"]

_LEASE_EXTEND_TIME_FACTOR = 0.8

_AIKO_LOG_LEVEL_LEASE = os.environ.get("AIKO_LOG_LEVEL_LEASE", "INFO")
_LOGGER = aiko.logger(__name__, log_level=_AIKO_LOG_LEVEL_LEASE)

class Lease:
    def __init__(
        self,
        lease_time,
        lease_uuid,
        lease_expired_handler=None,
        lease_extend_handler=None,
        automatic_extend=False):

        self.lease_time = lease_time
        self.lease_uuid = lease_uuid
        self.lease_expired_handler = lease_expired_handler
        self.lease_extend_handler = lease_extend_handler
        self.automatic_extend = automatic_extend

        event.add_timer_handler(self._lease_expired_timer, lease_time)
        if self.automatic_extend:
            extend_time = self.lease_time * _LEASE_EXTEND_TIME_FACTOR
            event.add_timer_handler(self.extend, extend_time)
        if _LOGGER.isEnabledFor(DEBUG):  # Don't expand debug message
            _LOGGER.debug(f"Lease created: {lease_uuid}: time={lease_time}")

    def extend(self, lease_time=None):
        if lease_time:
            self.lease_time = lease_time
        event.remove_timer_handler(self._lease_expired_timer)
        event.add_timer_handler(self._lease_expired_timer, self.lease_time)
        if self.lease_extend_handler:
            self.lease_extend_handler(self.lease_time, self.lease_uuid)
        if _LOGGER.isEnabledFor(DEBUG):  # Don't expand debug message
            _LOGGER.debug(
                f"Lease extended: {self.lease_uuid}, time={self.lease_time}")

    def _lease_expired_timer(self):
        event.remove_timer_handler(self._lease_expired_timer)
        if self.lease_expired_handler:
            self.lease_expired_handler(self.lease_uuid)
        if _LOGGER.isEnabledFor(DEBUG):  # Don't expand debug message
            _LOGGER.debug(f"Lease expired: {self.lease_uuid}")

    def terminate(self):
        event.remove_timer_handler(self._lease_expired_timer)
        if self.automatic_extend:
            event.remove_timer_handler(self.extend)
        if _LOGGER.isEnabledFor(DEBUG):  # Don't expand debug message
            _LOGGER.debug(f"Lease terminated: {self.lease_uuid}")



import shutil
from copy import deepcopy
from pathlib import Path
from typing import Optional

import ultralytics
import yaml

from highlighter.datasets.cropping import CropArgs
from highlighter.trainers._scaffold import TrainerType

__all__ = ["generate", "prepare_datasets", "train", "export"]


def generate(training_run_dir: Path, trainer: TrainerType):
    shutil.copy(Path(__file__).parent / "template.py", training_run_dir / "trainer.py")

    overrides_lookup = {
        TrainerType.YOLO_DET: {
            "model": "yolov8m.pt",
            "task": "detect",
        },
        TrainerType.YOLO_SEG: {
            "model": "yolov8m-seg.pt",
            # Ensures overlapping masks are merged, which is critical for accurate
            # segmentation in datasets with overlapping objects (default is True, but
            # explicitly set for clarity).
            "overlap_mask": True,
            # Controls mask resolution; 4 is the default and balances detail
            # with computational efficiency, suitable for most segmentation tasks.
            "mask_ratio": 4,
            "task": "segment",
        },
        TrainerType.YOLO_CLS: {
            "model": "yolov8m-cls.pt",
            # Adds regularization to prevent overfitting, which is more common
            # in classification tasks with large datasets; 0.1 is a modest starting point.
            "dropout": 0.1,
            # Classification often works with smaller images than detection/segmentation;
            # 224 is a common size (e.g., ImageNet), balancing detail and speed.
            "imgsz": 224,
            "task": "classify",
        },
    }

    overrides = overrides_lookup[trainer]

    default_cfg = dict(ultralytics.cfg.get_cfg())
    default_cfg.update(overrides)
    default_cfg["project"] = "runs"
    default_cfg["opset"] = 14
    default_cfg["format"] = "onnx"
    default_cfg["dynamic"] = False
    with (training_run_dir / "cfg.yaml").open("w") as f:
        yaml.dump(default_cfg, f)


def prepare_datasets(datasets):
    # When creating a training run in Highlighter the train split is required
    # but a user can supply either a test or dev set, or both. If not both we
    # duplicate the one that exists here
    if "test" not in datasets:
        datasets["test"] = deepcopy(datasets["dev"])
        datasets["test"].data_files_df.split = "test"
    if "dev" not in datasets:
        datasets["dev"] = deepcopy(datasets["test"])
        datasets["dev"].data_files_df.split = "dev"

    # Combine the Highlighter Datasets together because this is what the YoloWriter
    # expects
    combined_ds = datasets["train"]
    combined_ds.append([datasets["dev"], datasets["test"]])

    # Ultralytics name their dataset splits differently so
    # we need to map them
    #   their "val" is our "test"
    #   their "test" is our "dev"
    combined_ds.data_files_df.loc[combined_ds.data_files_df.split == "test", "split"] = "val"
    combined_ds.data_files_df.loc[combined_ds.data_files_df.split == "dev", "split"] = "test"
    return combined_ds


def _make_classify_artefact(onnx_filepath, cfg, crop_args) -> Path:
    d = dict(
        file_url=str(Path(onnx_filepath.absolute())),
        type="OnnxOpset14",
        inference_config=dict(
            type="classifier",
            code="BoxClassifier",
            machine_agent_type_id="d4787671-3839-4af9-9b34-a686faafbfae",
            parameters=dict(
                output_format="yolov8_cls",
                cropper=crop_args.model_dump(),
            ),
        ),
        training_config=cfg,
    )

    artefact_path = onnx_filepath.parent / "artefact.yaml"
    with artefact_path.open("w") as f:
        yaml.dump(d, f)

    return artefact_path


def _make_detect_segment_artefact(onnx_filepath, cfg) -> Path:
    output_format = "yolov8_seg" if cfg["task"] == "segment" else "yolov8_det"
    d = dict(
        file_url=str(Path(onnx_filepath.absolute())),
        type="OnnxOpset14",
        inference_config=dict(
            type="detector",
            code="Detector",
            machine_agent_type_id="29653174-8f45-440d-b75a-4ed0aa5fa6ff",
            parameters=dict(
                output_format=output_format,
            ),
        ),
        training_config=cfg,
    )

    artefact_path = onnx_filepath.parent / "artefact.yaml"
    with artefact_path.open("w") as f:
        yaml.dump(d, f)

    return artefact_path


def make_artefact(model, cfg, crop_args) -> Path:
    # Disable ultralyitcs' auto install of packages
    ultralytics.utils.checks.AUTOINSTALL = False
    onnx_filepath = Path(model.export(format="onnx", batch=1, dynamic=False, device=0))

    if cfg["task"] == "classify":
        return _make_classify_artefact(onnx_filepath, cfg, crop_args)
    elif cfg["task"] in ("detect", "segment"):
        return _make_detect_segment_artefact(onnx_filepath, cfg)


def train(cfg, training_run_dir: Path, crop_args: Optional[CropArgs]):
    model = ultralytics.YOLO(cfg["model"])

    ultralytics.settings.update({"datasets_dir": str(training_run_dir.absolute())})

    data_cfg_path = (training_run_dir / "datasets" / "data.yaml").absolute()
    with data_cfg_path.open("r") as f:
        data_cfg = yaml.safe_load(f)

    with data_cfg_path.open("w") as f:
        yaml.dump(data_cfg, f)

    if cfg["task"] == "classify":
        cfg["data"] = str(data_cfg_path.parent)
    else:
        cfg["data"] = str(data_cfg_path)

    model.train(**cfg)
    artefact_path = make_artefact(model, cfg, crop_args)

    return model, artefact_path.absolute()


def export(checkpoint, cfg, training_run_dir, crop_args, cfg_overrides={}):

    if isinstance(cfg, (str, Path)):
        with open(cfg, "r") as f:
            cfg = yaml.safe_load(f)

    cfg["model"] = checkpoint
    cfg.update(cfg_overrides)
    model = ultralytics.YOLO(cfg["model"])
    artefact_path = make_artefact(model, cfg, crop_args)

    return artefact_path.absolute()



# Description
# ~~~~~~~~~~~
# Hooks enable developers to flexibly extend the Aiko Services framework.
# A Hook is created within the framework and used by a third-party developer.
# For a given framework component ... the Hooks are a class variable
#
# Create a Hook within the framework ...
# - A component is composed with the "Hooks" interface
# - A new Hook is created via "self.add_hook(hook_name)"
# - Invoke "run_hook(hook_name)", which calls all provided "hook_functions()"
#
# Third-party developer extends the framework Hook ...
# - Provides "hook_function(hook_name, component, logger, variables)"
# - Add the hook function via "self.add_hook_handler(hook_name, hook_function)
#
# Hooks are only supported within the "component (Service)" infrastructure.
#
# Resource costs
# ~~~~~~~~~~~~~~
# CPU time used by run_hook() ...
# - 0x hook_handlers:  1 microsecond
# - 1x hook_handlers: 14 microseconds
# - 2x hook_handlers: 24 microseconds
#
# Usage: framework
# ~~~~~~~~~~~~~~~~
# NAME = "component.hook:version"  # if hook variables change, then bump version
#
# class HookTest(aiko.Actor):
#     def __init__(self, context):
#         self.add_hook(NAME)
#
#     def method_with_a_hook(self):
#         self.run_hook(NAME, lambda: {"variable", variable_value})
#
# Usage: third-party developer
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# add_hook_handler(NAME, self.hook_function)
#
# def hook_function(self, hook_name, component, logger, variables):
#     logger.debug(f"{hook_name} invoked for {component} with {variables}")
#
# Test
# ~~~~
# pytest ../tests/unit/test_hook.py
#
# To Do
# ~~~~~
# - Refactor Metrics to use Hooks for capturing CPU time and beyond !

from abc import abstractmethod
from collections import OrderedDict as OrderedDictCollections
from dataclasses import dataclass, field, is_dataclass
from typing import Any, Callable, Dict, OrderedDict as OrderedDictTyping

from aiko_services.main import *

__all__ = ["DEFAULT_HOOK", "Hook", "Hooks"]

ENABLED_DEFAULT = False

# --------------------------------------------------------------------------- #

@dataclass
class HookHandler:
    function: Callable[[object, object, dict], None]
    options: Dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        self.hash = hash(self.function) + hash(repr(self.options))

@dataclass
class Hook:
    name: str  # "component_name.hook_name:version"
    enabled: bool = ENABLED_DEFAULT
    handlers: OrderedDictTyping[str, HookHandler] =  \
                  field(default_factory=OrderedDictCollections)
    invoked: int = 0

class Hooks:
    Interface.default("Hooks", "aiko_services.main.hook.HooksImpl")

    @abstractmethod
    def add_hook(self, hook_name):
        pass

    @abstractmethod
    def add_hook_handler(self, hook_name, hook_function, hook_options=None):
        pass

    @abstractmethod
    def get_hook(self, hook_name):
        pass

    @abstractmethod
    def get_hooks(self):
        pass

    @abstractmethod
    def remove_hook(self, hook_name):
        pass

    @abstractmethod
    def remove_hook_handler(self, hook_name, hook_function):
        pass

    @abstractmethod
    def run_hook(self, hook_name):
        pass

    @abstractmethod
    def set_hook_enabled(self, hook_name, enabled_flag):
        pass

class HooksImpl(Hooks):
    hooks: Dict[str, Hook] = {}

    def add_hook(self, hook_name):
        if not self.get_hook(hook_name):
            self.get_hooks()[hook_name] = Hook(hook_name)

    def add_hook_handler(self, hook_name, hook_function, hook_options=None):
        hook = self.get_hook(hook_name)
        if not hook:
            raise RuntimeError(f"Hook {hook_name}: Does not exist")
        hook_options = hook_options if hook_options else {}
        hook_handler = HookHandler(hook_function, hook_options)
        hook.handlers[hook_handler.hash] = hook_handler
        hook.enabled = len(hook.handlers) > 0

    def get_hook(self, hook_name):
        if hook_name in HooksImpl.hooks:
            return HooksImpl.hooks[hook_name]
        else:
            return None

    def get_hooks(self):
        return HooksImpl.hooks

    def remove_hook(self, hook_name):
        hook = self.get_hook(hook_name)
        if not hook:
            raise RuntimeError(f"Hook {hook_name}: Does not exist")
        del self.get_hooks()[hook_name]

    def remove_hook_handler(self, hook_name, hook_function, hook_options=None):
        hook = self.get_hook(hook_name)
        if not hook:
            raise RuntimeError(f"Hook {hook_name}: Does not exist")
        hook_options = hook_options if hook_options else {}
        hook_handler = HookHandler(hook_function, hook_options)
        del hook.handlers[hook_handler.hash]
        hook.enabled = len(hook.handlers) > 0

    def run_hook(self, hook_name, variables=None):
        hook = self.get_hook(hook_name)
        if hook and hook.enabled:
            component = self
            logger = self.logger if hasattr(component, "logger") else None
            if not variables:
                variables = {}
            elif callable(variables):
                variables = variables()

            hook.invoked += 1
            for hook_handler in hook.handlers.values():
                options = hook_handler.options
                hook_handler.function(
                    hook_name, component, logger, variables, options)

    def set_hook_enabled(self, hook_name, enabled_flag):
        hook = self.get_hook(hook_name)
        if hook:
            hook.enabled = enabled_flag

# --------------------------------------------------------------------------- #

def hook_function(hook_name, component, logger, variables, hook_options=None):
    show = variables
    if hook_options and "show" in hook_options:
        show = {}
        for name_path in hook_options["show"]:
            names = name_path.split(".")
            value = variables

            for name in names:
                if isinstance(value, dict):
                    value = value[name]
                elif is_dataclass(value):
                    value = getattr(value, name, None)

            show[name] = value
    logger.info(f"Hook {hook_name}: {show}")

DEFAULT_HOOK = hook_function

# --------------------------------------------------------------------------- #



# Description
# ~~~~~~~~~~~
# Provide Service, Actor, PipelineElement and Pipeline instances with
# extensible data structures for storing common variable fields.
# This reduces the constructor "__init__()" arguments to just one.
# Providing some reasonable protection from breaking application
# code over time ... as the framework changes.
#
# Example
# ~~~~~~~
# In the following example, "actor_args()" creates a "ContextService"
# instance, which is used to populate the constructor "context" argument.
#
#     class Example(Actor):
#         def __init__(self, context):
#             context.get_implementation("Actor").__init__(self, context)
#
#     init_args = actor_args("example")
#     compose_instance(Example, init_args)
#     aiko.process.run()
#
# To Do
# ~~~~~
# - Provide get_parameter() and set_parameter() (using existing code) ?
#
# - Use "__file__" instead of "__name__" and ...
#   - Provide "__file__" --> "__name__" function to avoid "__main__"
#
# - Consider "protocol" revision versus "source code file" version ?
#   - Convenience function for automatically creating source code version ?
#
# - Provide custom __repr__() and __str__() ?
#
# - Provide add_tags() and remove_tags() (using existing code) ?
#   - Tags input is a list, but should tags internally be a dictionary ?

from abc import ABC
from dataclasses import dataclass, field
from typing import Dict, List

__all__ = {
    "Context", "Interface", "ServiceProtocolInterface", "ContextService",
    "ContextPipelineElement", "ContextPipeline"
    "service_args", "actor_args", "pipeline_element_args", "pipeline_args"
}

DEFAULT_PARAMETERS = {}
DEFAULT_PROTOCOL = "*"
DEFAULT_TAGS = []
DEFAULT_TRANSPORT = "mqtt"
DEFAULT_DEFINITION = ""
DEFAULT_DEFINITION_PATHNAME = ""
DEFAULT_STREAM_ID = 0
DEFAULT_FRAME_ID = 0

@dataclass
class Context:
    name: str = "<interface>"
    implementations: Dict[str, str] = field(default_factory=dict)

    def get_implementation(self, implementation_name):
        return self.implementations[implementation_name]

    def get_implementations(self):
        return self.implementations

    def get_name(self) -> str:
        return self.name

    def set_implementation(self, implementation_name, implementation):
        self.implementations[implementation_name] = implementation

    def set_implementations(self, implementations):
        self.implementations = implementations

class Interface(ABC):
    context = Context()

    @classmethod
    def default(cls, implementation_name, implementation):
        cls.context.set_implementation(implementation_name, implementation)

    @classmethod
    def get_implementations(cls):
        return cls.context.get_implementations()

class ServiceProtocolInterface(Interface):
    """Interface marker represents an Aiko Service implementing a protocol"""

# Service and Actor use the same data structure
@dataclass
class ContextService(Context):
    parameters: Dict[str, str] = field(default_factory=dict)
    protocol: str = DEFAULT_PROTOCOL
    tags: List[str] = field(default_factory=list)
    transport: str = DEFAULT_TRANSPORT

    def __post_init__(self):
        if self.name is None or not isinstance(self.name, str):
            raise ValueError(f"Service name must be a string: {self.name}")
        elif not len(self.name):
            raise ValueError(f"Service name must not be an empty string")
        if self.parameters is None:
            self.parameters = DEFAULT_PARAMETERS
        if self.protocol is None:
            self.protocol = DEFAULT_PROTOCOL
        if self.tags is None:
            self.tags = DEFAULT_TAGS
        if self.transport is None:
            self.transport = DEFAULT_TRANSPORT

    def get_parameters(self) -> Dict[str, str]:
        return self.parameters

    def get_protocol(self) -> str:
        return self.protocol

    def get_tags(self) -> List[str]:
        return self.tags

    def get_transport(self) -> str:
        return self.transport

    def set_protocol(self, protocol):
        self.protocol = protocol

@dataclass
class ContextPipelineElement(ContextService):
    definition: str = DEFAULT_DEFINITION
    pipeline: str = None  # TODO: Pipeline reference or None

    def __post_init__(self):
        self.name = self.name.lower()
        super().__post_init__()
        if self.definition is None:
            self.definition = DEFAULT_DEFINITION

    def get_definition(self) -> str:
        return self.definition

    def get_pipeline(self) -> str:
        return self.pipeline

@dataclass
class ContextPipeline(ContextPipelineElement):
    definition_pathname: str = DEFAULT_DEFINITION_PATHNAME
    graph_path: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.definition_pathname is None:
            self.definition_pathname = DEFAULT_DEFINITION_PATHNAME

    def get_definition_pathname(self) -> str:
        return self.definition_pathname

    def get_graph_path(self) -> str:
        return self.graph_path

def service_args(name, implementations=None,
    parameters=None, protocol=None, tags=None, transport=None):

    return {"context":
        ContextService(name, implementations,
            parameters, protocol, tags, transport)}

def actor_args(name, implementations=None,
    parameters=None, protocol=None, tags=None, transport=None):

    return service_args(name, implementations,
        parameters, protocol, tags, transport)

def pipeline_element_args(name, implementations=None,
    parameters=None, protocol=None, tags=None, transport=None,
    definition=None, pipeline=None):

    return {"context":
        ContextPipelineElement(name, implementations,
            parameters, protocol, tags, transport,
            definition, pipeline)}

def pipeline_args(name, implementations=None,
    parameters=None, protocol=None, tags=None, transport=None,
    definition=None, pipeline=None, definition_pathname=None,
    graph_path=None):

    return {"context":
        ContextPipeline(name, implementations,
            parameters, protocol, tags, transport,
            definition, pipeline, definition_pathname, graph_path)}



import json
from datetime import datetime
from uuid import UUID

__all__ = ["HLJSONEncoder"]


class HLJSONEncoder(json.JSONEncoder):

    def default(self, o):

        if isinstance(o, UUID):
            return str(o)
        elif isinstance(o, datetime):
            return o.isoformat()
        elif serialize_fn := getattr(o, "serialize", None):
            return serialize_fn()
        return super().default(o)

    def _encode_dict_keys(self, d):
        """
        Recursively traverses a nested dictionary and casts any UUID keys to strings.

        Args:
            d (dict): A dictionary that may contain UUID keys.

        Returns:
            dict: A new dictionary with UUID keys converted to strings.
        """
        if isinstance(d, dict):
            new_dict = {}
            for key, value in d.items():
                # Check if the key is a UUID, and if so, cast it to a string
                if isinstance(key, UUID):
                    new_key = str(key)
                else:
                    new_key = key

                # Recursively process the value if it's a nested dict or list
                if isinstance(value, dict):
                    new_dict[new_key] = self._encode_dict_keys(value)
                elif isinstance(value, list):
                    new_dict[new_key] = [
                        self._encode_dict_keys(item) if isinstance(item, dict) else item for item in value
                    ]
                else:
                    new_dict[new_key] = value
            return new_dict

        elif isinstance(d, (list, tuple)):
            return [self._encode_dict_keys(e) for e in d]
        return d

    def encode(self, obj):
        _obj = self._encode_dict_keys(obj)
        return super().encode(_obj)



from datetime import datetime
from typing import Dict, List, Tuple
from uuid import uuid4

from highlighter.agent.capabilities.base_capability import Capability, StreamEvent
from highlighter.client.base_models import DataFile

__all__ = ["PackDataFiles", "UnpackDataFiles"]


class PackDataFiles(Capability):

    class DefaultStreamParameters(Capability.DefaultStreamParameters):
        input_key: str
        content_type: str
        output_key: str = "data_files"
        output_empty_entities: bool = False

    @property
    def input_key(self):
        return self._get_parameter("input_key")[0]

    @property
    def content_type(self):
        return self._get_parameter("content_type")[0]

    @property
    def output_key(self):
        return self._get_parameter("output_key")[0]

    @property
    def output_empty_entities(self):
        return self._get_parameter("output_empty_entities")[0]

    def process_frame(self, stream, *args, **kwargs) -> Tuple[StreamEvent, Dict]:
        inputs = kwargs[self.input_key]

        if "timestamps" in stream.variables:
            timestamps = [datetime.fromtimestamp(ts) for ts in stream.variables["timestamps"]]
        else:
            rate, found = self.pipeline.pipeline_graph.nodes()[0].element.get_parameter("rate")
            if not found:
                raise ValueError("Cannot determine frame 'rate' from head node")

            frame_counter = getattr(self, "frame_counter", 0)
            timestamps = [1 / rate * (i + frame_counter) for i in range(len(inputs))]
            self.frame_counter = frame_counter + len(timestamps)

            # ToDo: Update the aiko fork when aiko starts putting `timestamps` in stream.variables
            #       the above code can be removed and we simply raise
            # raise ValueError("stream.variables must have 'timestamps' in order to create a DataFile object")

        data_files = [
            DataFile(recorded_at=ts, content=i, content_type=self.content_type, file_id=uuid4())
            for i, ts in zip(inputs, timestamps)
        ]
        result = {self.output_key: data_files}
        if self.output_empty_entities:
            result["entities"] = [{}] * len(data_files)

        return StreamEvent.OKAY, result


class UnpackDataFiles(Capability):

    class DefaultStreamParameters(Capability.DefaultStreamParameters):
        output_key: str

    @property
    def output_key(self):
        return self._get_parameter("output_key")[0]

    def process_frame(self, stream, data_files: List[DataFile], *args, **kwargs) -> Tuple[StreamEvent, Dict]:
        output = [d.content for d in data_files]
        return StreamEvent.OKAY, {self.output_key: output}



#!/usr/bin/env python3
#
# Usage
# ~~~~~
# ./lifecycle.py manager [client_count]
#
# ./lifecycle.py client client_id lifecycle_manager_topic
#
# Environment variables
# ~~~~~~~~~~~~~~~~~~~~~
# Fine-grained logging control ...
# - Show some low-level LifeCycle operations
#   - AIKO_LOG_LEVEL_LIFECYCLE: ERROR, WARNING, INFO, DEBUG
#
# To Do: LifeCycleManager
# ~~~~~~~~~~~~~~~~~~~~~~~
# - Implement "lifecycle_manager.delete(lifecycle_client_id)"
#
# - BUG: Fix "aiko_services/event.py" ...
#        With multiple timer handlers using the same handler function,
#        then remove_timer_handler() make remove the wrong one.
#        Need an identifier to specific exactly which handler instance
#
# - CRITICAL: *** Check all MQTT messages for excessive quantity ***
#   - Every LifeCycleClient is using ActorDiscovery/ServicesCache and
#     receiving notification on every other LifeCycleClient on the
#     Registrar topic "/out" ?!?
#
# - PERFORMANCE: Why does LifeCycleClient creation on AikoDashboard appear
#   so delayed / chunky ?  Where are the bottlenecks ?
#
# - BUG: LifeCycleManager and LifeCycleClient should handle degradation of
#   network (ConnectionState changes) and/or lease expiring.  Should LCM
#   recreate LCCs or terminate ?  Should LCC terminate ?
#
# - PRIORITY: Optionally replace ProcessManager with Ray to create Actor
#
# - Should be able to create LifeCycleClients as either separate processes
#   or within the same process.  When the LifeCycleClients are all running
#   in the same process ... that could the same or a different to the
#   LifeCycleManager.
#
# - The LifeCycleManager / LifeCycleClient can be generalized to be
#   one manager Actor creates other client Actor(s) and then uses
#   ECProducer / ECConsumer to listen to specified variable(s),
#   not just the "lifecycle" state
#
# - Use ProcessManager to monitor state of LifeCycleClient process
#   - Use "process_exit_handler" to manage unexpected process termination
#
# - Refactor Handshake out as more generally useful concept,
#   e.g as part of ServiceDiscovery or ActorDiscovery
#
# To Do: LifeCycleClient
# ~~~~~~~~~~~~~~~~~~~~~~
# - LifeCycleClient exits when its manager exits

from abc import abstractmethod
import click
import os
import time
from typing import Dict, List

from aiko_services.main import *
from aiko_services.main.transport import *
from aiko_services.main.utilities import *

CLIENT_SHELL_COMMAND = "./lifecycle.py"

_VERSION = 0

ACTOR_TYPE_LIFECYCLE_MANAGER = "lifecycle_manager"
PROTOCOL_LIFECYCLE_MANAGER =  \
    f"{SERVICE_PROTOCOL_AIKO}/{ACTOR_TYPE_LIFECYCLE_MANAGER}:{_VERSION}"

ACTOR_TYPE_LIFECYCLE_CLIENT = "lifecycle_client"
PROTOCOL_LIFECYCLE_CLIENT =  \
    f"{SERVICE_PROTOCOL_AIKO}/{ACTOR_TYPE_LIFECYCLE_CLIENT}:{_VERSION}"

_DELETION_LEASE_TIME_DEFAULT = 30  # seconds
_HANDSHAKE_LEASE_TIME_DEFAULT = 30  # seconds

_AIKO_LOG_LEVEL_LIFECYCLE = os.environ.get("AIKO_LOG_LEVEL_LIFECYCLE", "INFO")
_LOGGER = aiko.logger(__name__, log_level=_AIKO_LOG_LEVEL_LIFECYCLE)

# ---------------------------------------------------------------------------- #
# Over time, intend to utilise the Handshake concept more broadly
#
#class HandshakeLease(Lease):
#    def __init__(
#        self, lease_time, handshake_id, handler, lease_expired_handler):
#
#        super().__init__(
#            lease_time, handshake_id,
#            lease_expired_handler=lease_expired_handler)
#        self.handler = handler

class LifeCycleClientDetails:
    def __init__(self, client_id, topic_path, ec_consumer=None):
        self.client_id = client_id
        self.ec_consumer = ec_consumer
        self.topic_path = topic_path

class LifeCycleManager(ServiceProtocolInterface):
    Interface.default("LifeCycleManager",
        "aiko_services.main.lifecycle.LifeCycleManagerImpl")

    @abstractmethod
    def lcm_create_client(self, parameters=None):
        """Public method for creating clients

        Handles bookkeeping of clients, calling self._lcm_create_client
        for the creation step
        """

    @abstractmethod
    def lcm_delete_client(self, client_id):
        """Public method for deleting clients

        Handles bookkeeping of clients, calling self._lcm_delete_client
        for the deletion step.
        """

class LifeCycleManagerPrivate(Interface):
    Interface.default("LifeCycleManagerPrivate",
        "aiko_services.main.lifecycle.LifeCycleManagerImpl")

    @abstractmethod
    def _lcm_create_client(
        self, client_id, lifecycle_manager_topic, parameters):
        """Creation of a new client"""

    @abstractmethod
    def _lcm_delete_client(self, client_id, force=False):
        """Deletion of a client"""

    @abstractmethod
    def _lcm_get_clients(self) -> Dict[str, str]:
        """Getter for clients"""

    @abstractmethod
    def _lcm_get_handshaking_clients(self) -> List[int]:
        """Return IDs of clients that have been created,
           but that have not yet added themselves"""

    @abstractmethod
    def _lcm_lookup_client_state(self, client_id, client_state_key):
        """Lookup value from state of client"""

class LifeCycleManagerImpl(LifeCycleManager, LifeCycleManagerPrivate):
    def __init__(
            self,
            lifecycle_client_change_handler=None,
            ec_producer=None,
            client_state_consumer_filter="(lifecycle)",
            handshake_lease_time=_HANDSHAKE_LEASE_TIME_DEFAULT,
            deletion_lease_time=_DELETION_LEASE_TIME_DEFAULT):

        self.lcm_lifecycle_client_change_handler =  \
            lifecycle_client_change_handler
        self.lcm_client_count = 0
        self.lcm_ec_producer = ec_producer
        self.lcm_client_state_consumer_filter = client_state_consumer_filter
        self.lcm_deletion_lease_time = deletion_lease_time
        self.lcm_deletion_leases = {}
        self.lcm_handshake_lease_time = handshake_lease_time
        self.lcm_handshakes = {}
        self.lcm_lifecycle_clients = {}
        self.add_message_handler(
            self._lcm_topic_control_handler, self.topic_control)
        self.lcm_ec_producer.update("lifecycle_manager", {})  # TODO: Remove !

        if self.lcm_ec_producer is not None:
            self.lcm_ec_producer.update("lifecycle_manager_clients_active", 0)

    def lcm_create_client(self, parameters=None):
        if parameters == None:
            parameters = {}
        client_id = self.lcm_client_count
        self.lcm_client_count += 1
        self._lcm_create_client(client_id, self.topic_path, parameters)
        handshake = Lease(
            self.lcm_handshake_lease_time, client_id,
            lease_expired_handler=self._lcm_handshake_lease_expired_handler)
        self.lcm_handshakes[client_id] = handshake

    def lcm_delete_client(self, client_id):
        if client_id not in self.lcm_deletion_leases:
            self._lcm_delete_client(client_id)
            deletion_lease = Lease(
                self.lcm_deletion_lease_time, client_id,
                lease_expired_handler=self._lcm_deletion_lease_expired_handler)
            self.lcm_deletion_leases[client_id] = deletion_lease

    def _lcm_topic_control_handler(self, _aiko, topic, payload_in):
        command, parameters = parse(payload_in)

        if command == "add_client":
            lifecycle_client_topic_path = parameters[0]
            client_id = int(parameters[1])
            if client_id not in self.lcm_handshakes:
                _LOGGER.debug(f"LifeCycleClient {client_id} unknown")
            else:
                self.lcm_handshakes[client_id].terminate()
                del self.lcm_handshakes[client_id]
                _LOGGER.debug(f"LifeCycleClient {client_id} responded")

                class NullClass():
                    def null(self): pass

                do_discovery(NullClass,
                    ServiceFilter([lifecycle_client_topic_path],
                        "*", "*", "*", "*", "*"),
                    None, self._lcm_service_remove_handler)

                ec_consumer = ECConsumer(
                    self,
                    client_id, {},
                    f"{lifecycle_client_topic_path}/control",
                    self.lcm_client_state_consumer_filter)
                if self.lcm_lifecycle_client_change_handler:
                    ec_consumer.add_handler(self.lcm_lifecycle_client_change_handler)
                lifecycle_client_details = LifeCycleClientDetails(
                    client_id, lifecycle_client_topic_path, ec_consumer)
                self.lcm_lifecycle_clients[client_id] = lifecycle_client_details
                if self.lcm_ec_producer is not None:
                    self.lcm_ec_producer.update(
                        "lifecycle_manager_clients_active",
                        len(self.lcm_lifecycle_clients))
# TODO: This is a significant performance problem for large numbers of clients
                    self.lcm_ec_producer.update(
                        f"lifecycle_manager.{client_id}",
                        lifecycle_client_topic_path)

    def _lcm_service_remove_handler(self, service_details):
        lifecycle_client_topic_path = service_details[0]
        lifecycle_clients = list(self.lcm_lifecycle_clients.values())
        for lifecycle_client in lifecycle_clients:
            if lifecycle_client.topic_path == lifecycle_client_topic_path:
                if lifecycle_client.ec_consumer:
                    lifecycle_client.ec_consumer.terminate()
                    lifecycle_client.ec_consumer = None
                client_id = lifecycle_client.client_id

                if client_id in self.lcm_deletion_leases:
                    self.lcm_deletion_leases[client_id].terminate()
                    del self.lcm_deletion_leases[client_id]
                    _LOGGER.debug(f"LifeCycleClient {client_id} removed")

                del self.lcm_lifecycle_clients[client_id]
                if self.lcm_ec_producer is not None:
                    self.lcm_ec_producer.update(
                        "lifecycle_manager_clients_active",
                        len(self.lcm_lifecycle_clients))
                    self.lcm_ec_producer.remove(
                        f"lifecycle_manager.{client_id}")

            # When Service is removed from the Registrar, notify LCM Impl
                if self.lcm_lifecycle_client_change_handler:
                    self.lcm_lifecycle_client_change_handler(
                        client_id, "update", "lifecycle", "absent"
                    )

    def _lcm_deletion_lease_expired_handler(self, client_id):
        _LOGGER.debug(f"LifeCycleClient {client_id} deletion lease expired, force-deleting client")
        if client_id in self.lcm_deletion_leases:
            del self.lcm_deletion_leases[client_id]
        self._lcm_delete_client(client_id, force=True)

    def _lcm_handshake_lease_expired_handler(self, client_id):
        if client_id in self.lcm_handshakes:
            del self.lcm_handshakes[client_id]
        self._lcm_delete_client(client_id)
        _LOGGER.debug(f"LifeCycleClient {client_id} handshake failed")

    def _lcm_get_clients(self):
        clients = self.lcm_ec_producer.get("lifecycle_manager")
        if clients:
            clients = clients.copy()
            clients = { int(k): v for k, v in clients.items() }
        return clients

    def _lcm_get_handshaking_clients(self):
        return list(self.lcm_handshakes.keys())

    def _lcm_lookup_client_state(self, client_id, client_state_key):
        client_state_value = None
        client_details = self.lcm_lifecycle_clients.get(client_id)
        if client_details:
            if client_details.ec_consumer:
                client_state_value =  \
                    client_details.ec_consumer.cache.get(client_state_key)
        return client_state_value

# --------------------------------------------------------------------------- #

class LifeCycleManagerTest(Actor, LifeCycleManager):
    Interface.default("LifeCycleManagerTest",
        "aiko_services.main.lifecycle.LifeCycleManagerTestImpl")

class LifeCycleManagerTestImpl(LifeCycleManagerTest):
    def __init__(self, context, client_count):
        context.get_implementation("Actor").__init__(self, context)

        self.share = {
            "lifecycle": "ready",
            "log_level": get_log_level_name(_LOGGER),
            "source_file": f"v{_VERSION}⇒ {__file__}",
            "client_count": client_count
        }
        self.ec_producer = ECProducer(self, self.share)
        self.process_manager = ProcessManager()

        context.get_implementation("LifeCycleManager").__init__(self,
            self._lifecycle_client_change_handler, self.ec_producer)

        aiko.connection.add_handler(self._connection_state_handler)

    def _lcm_create_client(
        self, client_id, lifecycle_manager_topic, parameters):

        command = parameters
        arguments = ["client", str(client_id), lifecycle_manager_topic]
        self.process_manager.create(client_id, command, arguments)

    def _lcm_delete_client(self, client_id, force=False):
        self.process_manager.delete(client_id, kill=True)

    def _connection_state_handler(self, connection, connection_state):
        if connection.is_connected(ConnectionState.REGISTRAR):
            for client_count in range(self.share["client_count"]):
                lifecycle_client_id =  \
                    self.lcm_create_client(CLIENT_SHELL_COMMAND)
                time.sleep(0.01)

    def _lifecycle_client_change_handler(
        self, client_id, command, item_name, item_value):

        _LOGGER.debug(
            f"LifeCycleClient: {client_id}: {command} {item_name} {item_value}")

# ---------------------------------------------------------------------------- #

class LifeCycleClient(ServiceProtocolInterface):
    Interface.default("LifeCycleClient",
        "aiko_services.main.lifecycle.LifeCycleClientImpl")

class LifeCycleClientPrivate(Interface):
    Interface.default("LifeCycleClientPrivate",
        "aiko_services.main.lifecycle.LifeCycleClientImpl")

    @abstractmethod
    def _lcc_get_lifecycle_manager_topic(self):
        pass

    @abstractmethod
    def _lcc_lifecycle_manager_change_handler(self, service_details):
        pass

class LifeCycleClientImpl(LifeCycleClient, LifeCycleClientPrivate):
    def __init__(self,
        context, client_id, lifecycle_manager_topic, ec_producer):

        self.lcc_added_to_lcm = False
        self.lcc_client_id = client_id
        self.lcc_ec_producer = ec_producer
        self.lcc_ec_producer.update(
            "lifecycle_client.lifecycle_manager_topic", lifecycle_manager_topic)
        aiko.connection.add_handler(self._lcc_connection_handler)

    def _lcc_get_lifecycle_manager_topic(self):
        return self.lcc_ec_producer.get("lifecycle_client.lifecycle_manager_topic")

    def _lcc_connection_handler(self, connection, connection_state):
        if connection.is_connected(ConnectionState.REGISTRAR):
            if not self.lcc_added_to_lcm:
                lcm_topic_path = self._lcc_get_lifecycle_manager_topic()
                topic = f"{lcm_topic_path}/control"
                payload_out = "(add_client "                \
                              f"{self.topic_path} "  \
                              f"{self.lcc_client_id})"
                aiko.message.publish(topic, payload_out)
                self.lcc_added_to_lcm = True

                # Add handler for LifeCycleManager removal from registrar
                do_discovery(LifeCycleManagerImpl,
                    ServiceFilter([lcm_topic_path], "*", "*", "*", "*", "*"),
                    None, self._lcc_lifecycle_manager_change_handler)

    def _lcc_lifecycle_manager_change_handler(self, service_details):
        pass

# --------------------------------------------------------------------------- #

class LifeCycleClientTest(Actor, LifeCycleClient):
    Interface.default("LifeCycleClientTest",
        "aiko_services.main.lifecycle.LifeCycleClientTestImpl")

class LifeCycleClientTestImpl(LifeCycleClientTest):
    def __init__(self, context, client_id, lifecycle_manager_topic):
        context.get_implementation("Actor").__init__(self, context)

        self.share = {
            "lifecycle": "ready",
            "log_level": get_log_level_name(_LOGGER),
            "source_file": f"v{_VERSION}⇒ {__file__}",
            "client_id": client_id
        }
        self.ec_producer = ECProducer(self, self.share)

        context.get_implementation("LifeCycleClient").__init__(self,
            context, client_id, lifecycle_manager_topic, self.ec_producer)

# TODO: When scaling up to lots of LifeCycleClient, every LifeCycleClient
# receiving the ServiceDetails for every other LifeCycleClient is too much.
# - Registrar should provide a filter that limits to the requested topic path
#   Use below to be notified only when that Service is added or removed
# - Could also filter on the LifeCycleManager protocol to limit results
#
#       filter = ServiceFilter(
#           [lifecycle_manager_topic], "*", "*", "*", "*", "*")
#       self.actor_discovery = ActorDiscovery(service)
#       self.actor_discovery.add_handler(
#           self._lifecycle_manager_change_handler, filter)
#
#   def _lifecycle_manager_change_handler(self, command, service_details):
#       if command == "remove":
#           self.delete()

# ---------------------------------------------------------------------------- #

@click.group()

def main():
    pass

@main.command(help="LifeCycleManager Actor")
@click.argument("client_count", default=1)

def manager(client_count):
    tags = ["ec=true"]  # TODO: Add ECProducer tag before add to Registrar
    init_args = actor_args(ACTOR_TYPE_LIFECYCLE_MANAGER,
        protocol=PROTOCOL_LIFECYCLE_MANAGER, tags=tags)
    init_args["client_count"] = client_count
    lifecycle_manager = compose_instance(LifeCycleManagerTestImpl, init_args)
    aiko.process.run()

@main.command(help="LifeCycleClient Actor")
@click.argument("client_id", default=None)
@click.argument("lifecycle_manager_topic", default=None)

def client(client_id, lifecycle_manager_topic):
    tags = ["ec=true"]  # TODO: Add ECProducer tag before add to Registrar
    name = f"{ACTOR_TYPE_LIFECYCLE_CLIENT}_{client_id}"
    init_args = actor_args(name, protocol=PROTOCOL_LIFECYCLE_CLIENT, tags=tags)
    init_args["client_id"] = client_id
    init_args["lifecycle_manager_topic"] = lifecycle_manager_topic
    life_cycle_client = compose_instance(LifeCycleClientTestImpl, init_args)
    aiko.process.run()

if __name__ == "__main__":
    main()

# ---------------------------------------------------------------------------- #



# Description
# ~~~~~~~~~~~
# Distributed Component that can be discovered and processes messages.
# Defines a Service that runs within a Process.
#
# Example
# ~~~~~~~
# from abc import abstractmethod
# from aiko_services.main import *
#
# class ServiceTest(Service):
#     Interface.default("ServiceTest", "__main__.ServiceTestImpl")
#
#     @abstractmethod
#     def test(self):
#         pass
#
# class ServiceTestImpl(ServiceTest):
#     def __init__(self, context):
#         context.get_implementation("Service").__init__(self, context)
#
#     def test(self):
#         print("ServiceTestImpl.test() invoked")
#
# protocol = f"{SERVICE_PROTOCOL_AIKO}/service_test:0"
# init_args = service_args("service_test", protocol)
# service_test = compose_instance(ServiceTestImpl, init_args)
# service_test.test()
# aiko.process.run()
#
# Notes
# ~~~~~
# topic_path = ServiceTopicPath(
#     "namespace",
#     "host",
#     "process_id",
#     "service_id")
#
# service_fields = ServiceFields(
#     "topic_path",
#     "name",
#     ServiceProtocol(SERVICE_PROTOCOL_AIKO, "test", "0"),
#     "transport",
#     "owner",
#     "tags")
#
# service_fields.topic_path = topic_path
#
# To Do
# ~~~~~
# - BUG: Provide filtered Services to "service_change_handler"
#
# - Service Hooks should be optional, whilst Actor Hooks are baked in
#
# - Consider using @dataclass, as per "pipeline.py"
#
# - If a Service is created after the Process has found the Registrar,
#   then ensure that the Service is added to the Registrar.
# - If a Service is terminated, ensure that it is removed from the Registrar.
#
# - Default topic_in_handler(): "message --> function call"
#   - Also provide remote proxy for "function call --> message"
#
# - Add Component.name (?) --> Service.name --> Actor.name
#
# - Consolidate Sevices Eventual Consistency and Registrar "ServicesCache"
#
# - Move ServiceImpl2 to tutorials or examples
#
# - Define ServiceProtocol: registrar.py, service.py (and more ?)
#   - All Services define their own static ServiceProtocol
#
# - Document ServiceFields, ServiceTags and ServiceTopicPath
#   - Document: Namespace, Name, Protocol, Transport and Tags
#
# - Consider extending ServiceFields to provide share.py:ServiceFilter
#
# - ServiceFields
#   - Parser and generator (for presentation)
#   - Validation for each setter
#       @name.setter
#       def name(self, value):
#           self._name = value
#
# - ServiceFilter
#   - Generator (for presentation)
#
# - ServiceTags
#   - Parser and generator (for presentation)
#   - Validation for each setter
#
# - ServiceTopicPath
#   - Parser and generator (for presentation)
#   - Validation for each setter

from abc import abstractmethod
from collections import OrderedDict  # All OrderedDict operations are O(1)
import time

from aiko_services.main import *

__all__ = [
    "ServiceFields", "ServiceFilter", "ServiceProtocol",
    "ServiceTags", "ServiceTopicPath", "Services",
    "Service", "ServiceImpl"
]

class ServiceProtocol:
    def __init__(self, url_prefix, name, version):
        self._url_prefix = url_prefix
        self._name = name
        self._version = version

    def __repr__(self):
        return f"{self._url_prefix}/{self._name}:{self._version}"

    @property
    def url_prefix(self):
        return self._url_prefix

    @url_prefix.setter
    def url_prefix(self, value):
        self._url_prefix = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, value):
        self._version = value

# class ServiceField:  # TODO: Support integer index plus string name
#   TOPIC = "TOPIC"          # 0
#   NAME = "NAME"            # 1
#   PROTOCOL = "PROTOCOL"    # 2
#   TRANSPORT = "TRANSPORT"  # 3
#   OWNER = "OWNER"          # 4
#   TAGS = "TAGS"            # 5

#   fields = [TOPIC, NAME, PROTOCOL, TRANSPORT, OWNER, TAGS]

class ServiceFields:
    def __init__(self, topic_path, name, protocol, transport, owner, tags):
        self._topic_path = topic_path
        self._name = name
        self._protocol = protocol
        self._transport = transport
        self._owner = owner
        self._tags = tags

    def __repr__(self):
        return f"{self._topic_path}, {self._name}, "  \
               f"{self._protocol}, {self._transport}, "      \
               f"{self._owner}, {self._tags}"

    @property
    def topic_path(self):
        return self._topic_path

    @topic_path.setter
    def topic_path(self, value):
        self._topic_path = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def protocol(self):
        return self._protocol

    @protocol.setter
    def protocol(self, value):
        self._protocol = value

    @property
    def transport(self):
        return self._transport

    @transport.setter
    def transport(self, value):
        self._transport = value

    @property
    def owner(self):
        return self._owner

    @owner.setter
    def owner(self, value):
        self._owner = value

    @property
    def tags(self):
        return self._tags

    @tags.setter
    def tags(self, value):
        self._tags = value

class ServiceFilter:
    @classmethod
    def with_topic_path(cls, topic_path="*",
        name="*", protocol="*", transport="*", owner="*", tags="*"):

        topic_paths = topic_path if topic_path == "*" else [topic_path]
        return ServiceFilter(
            topic_paths, name, protocol, transport, owner, tags)

    def __init__(self, topic_paths="*",
        name="*", protocol="*", transport="*", owner="*", tags="*"):

        self.topic_paths = topic_paths
        self.name = name
        self.protocol = protocol
        self.transport = transport
        self.owner = owner
        self.tags = tags

    def __repr__(self):
        return f"{self.topic_paths}, {self.name}, "    \
               f"{self.protocol}, {self.transport}, "  \
               f"{self.owner}, {self.tags}"

    def summary(self):
        summary  = "" if self.topic_paths == "*" else f", {self.topic_paths}"
        summary += "" if self.name        == "*" else f", {self.name}"
        summary += "" if self.protocol    == "*" else f", {self.protocol}"
        summary += "" if self.transport   == "*" else f", {self.transport}"
        summary += "" if self.owner       == "*" else f", {self.owner}"
        summary += "" if self.tags        == "*" else f", {self.tags}"
        return summary[2:] if len(summary) else "any"

class ServiceTags:  # TODO: Dictionary of keyword / value pairs
    @classmethod
    def get_tag_value(cls, key, tags):
        tags = ServiceTags.parse_tags(tags)
        return tags.get(key)

    @classmethod
    def match_tags(cls, service_tags, match_tags):
        return all([tag in service_tags for tag in match_tags])

    @classmethod
    def parse_tags(cls, tags_list):
        tags = {}
        for tag in tags_list:
            key, value = tag.split("=")
            tags[key] = value
        return tags

class ServiceTopicPath:
    @classmethod
    def parse(cls, topic_path):
        try:
            namespace, hostname, process_id, service_id = topic_path.split("/")
            return ServiceTopicPath(namespace, hostname, process_id, service_id)
        except ValueError:
            return None

    @classmethod
    def topic_paths(cls, topic_path):
        process_topic_path = None
        service_topic_path = ServiceTopicPath.parse(topic_path)
        if service_topic_path:
            process_topic_path = service_topic_path.topic_path_process
        return process_topic_path, str(service_topic_path)


    def __init__(self, namespace, hostname, process_id=0, service_id=0):
        self._namespace = namespace
        self._hostname = hostname
        self._process_id = process_id
        self._service_id = service_id

    def __repr__(self):
        return f"{self.topic_path_process}/{self._service_id}"

    @property
    def namespace(self):
        return self._namespace

    @namespace.setter
    def namespace(self, value):
        self._namespace = value

    @property
    def hostname(self):
        return self._hostname

    @hostname.setter
    def hostname(self, value):
        self._hostname = value

    @property
    def process_id(self):
        return self._process_id

    @process_id.setter
    def process_id(self, value):
        self._process_id = value

    @property
    def service_id(self):
        return self._service_id

    @service_id.setter
    def service_id(self, value):
        self._service_id = value

    @property
    def terse(self):
        topic_path = str(self)
        if len(topic_path) > 26:
            namespace = self._namespace[0:4]
            if len(namespace) < len(self._namespace):
                namespace += "+"
            hostname = self._hostname[0:8]
            if len(hostname) < len(self._hostname):
                hostname += "+"
            process_id = self._process_id
            service_id = self._service_id
            topic_path = f"{namespace}/{hostname}/{process_id}/{service_id}"
        return topic_path

    @property
    def topic_path_process(self):
        return f"{self._namespace}/{self._hostname}/{self._process_id}"

# --------------------------------------------------------------------------- #
# services dict: Key: Process topic_path --> Value: Per Processs Services dict

class ServicesIterator:
    def __init__(self, services):
        self._services = services
        self._process_iterator = iter(self._services)
        self.iterate_process()

    def iterate_process(self):
        process_topic_path = self._process_iterator.__next__()
        self._process_services = self._services[process_topic_path]
        self._service_iterator = iter(self._process_services)

    def __next__(self):
        try:
            service_topic_path = self._service_iterator.__next__()
        except StopIteration:
            self.iterate_process()
            service_topic_path = self._service_iterator.__next__()
        return self._process_services[service_topic_path]

class Services:
    def __init__(self):
        self._count = 0
        self._services = OrderedDict()

    def __iter__(self):
        return ServicesIterator(self._services) if self._services else iter([])

    def __str__(self):
        return "\n".join(self.get_topic_paths())

    def add_service(self, topic_path, service_details):
        process_topic_path, service_topic_path =  \
            ServiceTopicPath.topic_paths(topic_path)
        if process_topic_path:
            if not process_topic_path in self._services:
                index = "protocol" if isinstance(service_details, dict) else 2
                order = service_details[index] != REGISTRAR_PROTOCOL
                self._services.update({process_topic_path: {}})
                self._services.move_to_end(process_topic_path, last=order)
            process_services = self._services[process_topic_path]
            if not service_topic_path in process_services:
                process_services[service_topic_path] = service_details
                self._count += 1

    def copy(self):
        clone = Services()
        clone._services = self._services.copy()
        return clone

    @property
    def count(self):
        return self._count  # TODO: Why is "len(self._services)" incorrect ?

    def filter_services(self, filter):
        results = self.filter_by_topic_paths(filter.topic_paths)
        results = self.filter_by_attributes(filter, services=results)
        return results

# TODO: Make this a more general "filter" that is compatible with
#       "aiko_services/share.py:_filter_compare()" and keep them together
# TODO: Note this code is copied from "aiko_services/registrar.py"
#       The registrar should also use general filters
# services dict: Key: Process topic_path --> Value: Per Processs Services dict

    def filter_by_attributes(self, filter, services=None):
        services = services._services if services else self._services
        results = Services()

        for process_topic_path, process_services in services.items():
            for service_topic, service_details in process_services.items():
            # TODO: Consolidate into consistent ServiceDetails() data structure
                if isinstance(service_details, dict):
                    name = service_details["name"]
                    protocol = service_details["protocol"]
                    transport = service_details["transport"]
                    owner = service_details["owner"]
                    tags = service_details["tags"]
                else:
                    name = service_details[1]
                    protocol = service_details[2]
                    transport = service_details[3]
                    owner = service_details[4]
                    tags = service_details[5]
                matches = True
                if filter.name != "*":
                    if filter.name != name:
                        matches = False
                if filter.protocol != "*":
                    if filter.protocol != protocol:
                        matches = False
                if filter.transport != "*":
                    if filter.transport != transport:
                        matches = False
                if filter.owner != "*":
                    if filter.owner != owner:
                        matches = False
                if filter.tags != "*":
                    if not ServiceTags.match_tags(tags, filter.tags):
                        matches = False
                if matches:
                    results.add_service(service_topic, service_details)
        return results

# TODO: Currently unused
# TODO: Update to use ServiceFields.name, rather than "actor=name" tag
#   def filter_by_name(self, name):
#       services = {}
#       topic_path = ".".join(name.split(".")[:-1])
#       if topic_path in self._services.keys():
#           service = self._services[topic_path]
#           match_tag = f"actor={name}"  # TODO: service.name
#           if ServiceTags.match_tags(service[4], [match_tag]):
#               services[name] = service
#       return services

    def filter_by_topic_paths(self, topic_paths):
        if topic_paths == "*":
            results = self
        else:
            results = Services()
            for topic_path in topic_paths:
                process_topic_path, _ = ServiceTopicPath.topic_paths(topic_path)
                if process_topic_path in self._services:
                    process_services = self._services[process_topic_path]
                    if topic_path in process_services:
                        results.add_service(
                            topic_path, process_services[topic_path])
        return results

    def get_process_services(self, process_topic_path):
        if process_topic_path in self._services:
            return self._services[process_topic_path].keys()
        return []

    def get_service(self, topic_path):
        process_topic_path, service_topic_path =  \
            ServiceTopicPath.topic_paths(topic_path)
        if process_topic_path in self._services:
            process_services = self._services[process_topic_path]
            if service_topic_path in process_services:
                return process_services[service_topic_path]
        return None

    def get_topic_paths(self):
        topic_paths = []
        for _, process_services in self._services.items():
            topic_paths.extend(list(process_services.keys()))
        return topic_paths

    def remove_service(self, topic_path):
        process_topic_path, service_topic_path =  \
            ServiceTopicPath.topic_paths(topic_path)
        if process_topic_path in self._services:
            process_services = self._services[process_topic_path]
            if service_topic_path in process_services:
                del process_services[service_topic_path]
                self._count -= 1
            if len(process_services) == 0:
                del self._services[process_topic_path]

# --------------------------------------------------------------------------- #

class Service(ServiceProtocolInterface, Hooks):  # TODO: Make Hooks be optional
    Interface.default("Service", "aiko_services.main.service.ServiceImpl")

    @abstractmethod
    def add_message_handler(self, message_handler, topic, binary=False):
        pass

    @abstractmethod
    def remove_message_handler(self, message_handler, topic):
        pass

    @abstractmethod
    def registrar_handler_call(self, action, registrar):
        pass

    @abstractmethod
    def run(self):
        pass

    @abstractmethod
    def set_registrar_handler(self, registrar_handler):
        pass

    @abstractmethod
    def stop(self):
        pass

    @abstractmethod
    def add_tags(self, tags):
        pass

    @abstractmethod
    def add_tags_string(self, tags_string):
        pass

    @abstractmethod
    def get_tags_string(self):
        pass

# --------------------------------------------------------------------------- #

class ServiceImpl(Service):
    def __init__(self, context):

    # TODO: Move name, protocol, tags, topic_path, transport into ServiceFields
        self.time_started = time.monotonic()
        self.name = context.name
        self.protocol = context.protocol
        self._tags = context.tags
        self.transport = context.transport
        aiko.process.add_service(self)  # Initializes service_id and topic_path

        self._registrar_handler_function = None
        self.topic_control = f"{self.topic_path}/control"
        self.topic_in = f"{self.topic_path}/in"
        self.topic_log = f"{self.topic_path}/log"
        self.topic_out = f"{self.topic_path}/out"
        self.topic_state = f"{self.topic_path}/state"

    def add_message_handler(self, message_handler, topic, binary=False):
        aiko.process.add_message_handler(message_handler, topic, binary)

    def remove_message_handler(self, message_handler, topic):
        aiko.process.remove_message_handler(message_handler, topic)

    def registrar_handler_call(self, action, registrar):
        if self._registrar_handler_function:
            self._registrar_handler_function(action, registrar)

    def run(self):
        raise SystemExit("Unimplemented: Currently only supported by Actor")

    def set_registrar_handler(self, registrar_handler):
        self._registrar_handler_function = registrar_handler

    def stop(self):
        aiko.process.terminate()

    def add_tags(self, tags):
        for tag in tags:
            if not ServiceTags.match_tags(self._tags, [tag]):
                self._tags.append(tag)

    def add_tags_string(self, tags_string):
        if tags_string:
            tags = tags_string.split(",")
            self.add_tags(tags)

    def get_tags_string(self):
        return ' '.join([str(tag) for tag in self._tags])

# --------------------------------------------------------------------------- #



import json
import os
import time
from datetime import datetime, timezone
from pathlib import Path
from queue import Empty, Queue
from tempfile import mkdtemp
from threading import Thread
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from uuid import UUID

import aiko_services as aiko
import gql
import yaml
from aiko_services.main import aiko as aiko_main
from aiko_services.main import generate, parse

from highlighter import MachineAgentVersion
from highlighter.agent.capabilities.base_capability import StreamEvent
from highlighter.client.agents import create_pipeline_instance, update_pipeline_instance
from highlighter.client.base_models.base_models import UserType
from highlighter.client.gql_client import HLClient
from highlighter.client.tasks import (
    Task,
    TaskStatus,
    lease_task,
    lease_tasks_from_steps,
    update_task,
)
from highlighter.core.logging import get_default_logger

__all__ = [
    "HLAgent",
    "set_mock_aiko_messager",
    "SExpression",
]

logger = get_default_logger(__name__)


TASK_POLLING_PERIOD_SECONDS = 5
TASK_LEASE_DURATION_SECONDS = 60
# The stream timeout counts from the most recent process_frame call in a pipeline
STREAM_TIMEOUT_GRACE_TIME_SECONDS = 20
# The pipeline timeout counts from the most recent frame returned on queue_response
PIPELINE_TIMEOUT_SECONDS = 60
TASK_RE_LEASE_EXPIRY_BUFFER_SECONDS = 20


class SExpression:

    @staticmethod
    def encode(cmd: Optional[str], parameters: Union[Dict, List, Tuple]) -> str:
        if cmd:
            return generate(cmd, parameters)
        else:
            return generate(parameters[0], parameters[1:])

    @staticmethod
    def decode(s: str) -> Any:
        return parse(s)


def load_pipeline_definition(path) -> Dict:
    path = Path(path)
    suffix = path.suffix

    if suffix in (".json",):
        with path.open("r") as f:
            pipeline_def = json.load(f)
    elif suffix in (".yml", ".yaml"):
        with path.open("r") as f:
            pipeline_def = yaml.safe_load(f)
    else:
        raise NotImplementedError(
            f"Unsupported pipeline_definition file, '{path}'." " Expected .json|.yml|.yaml"
        )

    def remove_dict_keys_starting_with_a_hash(data):
        if isinstance(data, dict):
            # Create a new dictionary excluding keys starting with "#"
            return {
                key: remove_dict_keys_starting_with_a_hash(value)
                for key, value in data.items()
                if not key.startswith("#")
            }
        elif isinstance(data, list):
            # If the item is a list, recursively apply the function to each element
            return [remove_dict_keys_starting_with_a_hash(item) for item in data]
        else:
            # If the item is neither a dict nor a list, return it as-is
            return data

    pipeline_def = remove_dict_keys_starting_with_a_hash(pipeline_def)
    return pipeline_def


class HLPipelineImpl(aiko.PipelineImpl):

    def __init__(self, context):
        self.hl_task_status_info = {
            "state": aiko.StreamState.RUN,
            "message": "",
        }
        self.on_agent_error: Optional[Callable] = None
        self.on_agent_stop: Optional[Callable] = None
        self.on_before_process_frame: Optional[Callable] = None
        self.on_after_process_frame: Optional[Callable] = None
        super().__init__(context)

    def _process_stream_event(self, element_name, stream, stream_event, diagnostic, in_destroy_stream=False):
        # class StreamState:
        #    ERROR =   -2  # Don't generate new frames and ignore queued frames
        #    STOP  =   -1  # Don't generate new frames and process queued frames
        #    RUN   =    0  # Generate new frames and process queued frames

        most_recent_state = self.hl_task_status_info["state"]

        if stream_event == aiko.StreamEvent.ERROR:
            if aiko.StreamState.ERROR < most_recent_state:
                self.hl_task_status_info["state"] = aiko.StreamState.ERROR
                self.hl_task_status_info["message"] = diagnostic
            if self.on_agent_error is not None:
                self.on_agent_error(self)

        elif stream_event == aiko.StreamEvent.STOP:
            if aiko.StreamState.STOP < most_recent_state:
                self.hl_task_status_info["state"] = aiko.StreamState.STOP
                self.hl_task_status_info["message"] = diagnostic
            if self.on_agent_stop is not None:
                self.on_agent_stop(self)

        return super()._process_stream_event(
            element_name, stream, stream_event, diagnostic, in_destroy_stream=in_destroy_stream
        )

    def process_frame(self, stream_dict, frame_data) -> bool:
        try:
            if self.on_before_process_frame is not None:
                self.on_before_process_frame(self)
        except Exception as e:
            # TODO: This return value doesn't gel with aiko.PipelineImpl.process_frame
            return aiko.StreamEvent.ERROR, {"diagnostic": e}

        result = super().process_frame(stream_dict, frame_data)

        try:
            if self.on_after_process_frame is not None:
                self.on_after_process_frame(self)
        except Exception as e:
            # TODO: This return value doesn't gel with aiko.PipelineImpl.process_frame
            return aiko.StreamEvent.ERROR, {"diagnostic": e}

        return result


aiko.PipelineImpl = HLPipelineImpl


def _validate_uuid(s):
    try:
        u = UUID(s)
        return u
    except Exception as e:
        return None


def _validate_path(s) -> Optional[Path]:
    try:
        p = Path(s)
        if p.exists():
            return p
    except Exception as e:
        return None


class HLAgent:

    def __init__(
        self,
        pipeline_definition: Union[str, dict, os.PathLike],
        name: Optional[str] = "agent",
        dump_definition: Optional[os.PathLike] = None,
    ):
        if dump_definition is not None:
            pipeline_path = Path(dump_definition)
        else:
            pipeline_path = Path(mkdtemp()) / "pipeline_def.json"

        self.machine_agent_version_id = None

        if pipeline_definition_path := _validate_path(pipeline_definition):
            definition_dict = load_pipeline_definition(pipeline_definition_path)
            if name is None:
                name = pipeline_definition_path.name

        elif def_uuid := _validate_uuid(pipeline_definition):
            result = HLClient.get_client().machine_agent_version(
                return_type=MachineAgentVersion,
                id=str(def_uuid),
            )
            definition_dict = result.hl_serving_agent_definition
            name = result.title
            self.machine_agent_version_id = def_uuid
        elif isinstance(pipeline_definition, dict):
            if name is None:
                raise ValueError(
                    "If pipeline_definition is a dict you must provide the 'name' arg to HLAgent.__init__"
                )
            definition_dict = pipeline_definition

        else:
            if Path(pipeline_definition).suffix not in (".json", ".yml", ".yaml"):
                raise SystemExit(f"pipeline_definition '{pipeline_definition}' path does not exist")
            else:
                raise SystemExit(f"pipeline_definition '{pipeline_definition}' id does not exist")

        self._dump_definition(definition_dict, pipeline_path)

        parsed_definition = aiko.PipelineImpl.parse_pipeline_definition(pipeline_path)

        init_args = aiko.pipeline_args(
            name,
            protocol=aiko.PROTOCOL_PIPELINE,
            definition=parsed_definition,
            definition_pathname=pipeline_path,
        )
        pipeline = aiko.compose_instance(aiko.PipelineImpl, init_args)

        self.pipeline = pipeline
        self.pipeline_definition = parsed_definition

    def get_data_source_capabilities(self) -> List[aiko.source_target.DataSource]:
        data_source_elements = [
            node
            for node in self.pipeline.pipeline_graph.nodes()
            if isinstance(node.element, aiko.source_target.DataSource)
        ]
        return data_source_elements

    def _dump_definition(self, pipeline_def: Dict, path: Path):
        with path.open("w") as f:
            json.dump(pipeline_def, f, sort_keys=True, indent=2)

    def run_in_foreground(self, mqtt_connection_required=False):
        self.pipeline.run(mqtt_connection_required=mqtt_connection_required)

    def run_in_thread(self, mqtt_connection_required=False) -> Thread:
        thread = Thread(
            target=self.run_in_foreground,
            daemon=True,
            kwargs={"mqtt_connection_required": mqtt_connection_required},
        )
        thread.start()
        return thread

    def stop(self):
        self.pipeline.stop()
        while aiko.event.event_loop_running:
            time.sleep(0.01)

    def get_head_data_source_capability(self):
        head_capability = [x for x in self.pipeline.pipeline_graph][0]
        if hasattr(head_capability.element, "_is_data_source"):
            return head_capability
        return None

    def set_callback(self, callback_name: str, callback: Callable):
        setattr(self.pipeline, callback_name, callback)

    def process_frame(self, frame_data, stream_id=0, frame_id=0) -> bool:
        stream = {
            "stream_id": stream_id,
            "frame_id": frame_id,
        }
        return self.pipeline.process_frame(stream, frame_data)

    def poll_for_tasks_loop(self, step_id: Union[str, UUID], allow_non_machine_user: bool = False):
        current_user = HLClient.get_client().current_user(return_type=UserType)
        if current_user.machine_agent_version_id is None and not allow_non_machine_user:
            raise RuntimeError(
                "Running agent as non-machine user. To run the agent as a machine user, use "
                "`hl agent create-token` and set HL_WEB_GRAPHQL_API_TOKEN with the returned value before running `hl agent start. "
                "To run the agent as the current user, pass `allow_non_machine_user=True`."
            )

        step_id = UUID(step_id)

        # Report running agent to hl web
        self.pipeline_instance_id = create_pipeline_instance(
            str(self.machine_agent_version_id),
            str(step_id),
        )
        try:
            while True:
                update_pipeline_instance(self.pipeline_instance_id, status="RUNNING")
                tasks = lease_tasks_from_steps(
                    HLClient.get_client(),
                    [step_id],
                    lease_sec=TASK_LEASE_DURATION_SECONDS,
                    count=1,
                )
                for task in tasks:
                    self._process_task(task)
                if len(tasks) == 0:
                    time.sleep(TASK_POLLING_PERIOD_SECONDS)
        except KeyboardInterrupt as e:
            # SIGINT received, agent stopped but not failed
            try:
                update_pipeline_instance(self.pipeline_instance_id, status="STOPPED")
            except gql.transport.exceptions.TransportAlreadyConnected:
                # without asyncio
                client = HLClient.get_client()
                client._async = False
                update_pipeline_instance(self.pipeline_instance_id, status="STOPPED")
            raise e
        except Exception as e:
            update_pipeline_instance(self.pipeline_instance_id, status="FAILED", message=str(e))
            raise e

    def _process_task(self, task: Task):
        logger.info(f"Processing task {task.id}")
        stream_id = task.id
        stream_result_queue = Queue()
        self.pipeline.create_stream(
            stream_id,
            parameters=task.parameters,
            queue_response=stream_result_queue,
            grace_time=STREAM_TIMEOUT_GRACE_TIME_SECONDS,
        )
        while True:
            try:
                stream_info, frame_data = stream_result_queue.get(timeout=PIPELINE_TIMEOUT_SECONDS)
            except Empty:
                raise ValueError(
                    "Timeout: Pipeline has not produced a value for more than "
                    f"{PIPELINE_TIMEOUT_SECONDS} seconds"
                )
            if stream_info["state"] == aiko.StreamEvent.STOP:
                logger.info(f"Completed task {task.id}")
                update_task(
                    client=HLClient.get_client(),
                    task_id=task.id,
                    status=TaskStatus.SUCCESS,
                )
                break
            if stream_info["state"] == aiko.StreamEvent.ERROR:
                logger.error(f"Error in task {task.id}: {frame_data.get('diagnostic')}")
                update_task(
                    client=HLClient.get_client(),
                    task_id=task.id,
                    status=TaskStatus.FAILED,
                    message=frame_data.get("diagnostic"),
                )
                break
            if (
                task.leased_until.timestamp()
                < datetime.now(timezone.utc).timestamp() + TASK_RE_LEASE_EXPIRY_BUFFER_SECONDS
            ):
                # Re-lease task
                logger.info(f"Extending lease for task {task.id}")
                task = lease_task(
                    client=HLClient.get_client(),
                    task_id=task.id,
                    lease_sec=TASK_LEASE_DURATION_SECONDS,
                )


def set_mock_aiko_messager():
    # ToDo: Chat with Andy about if this is a requirement. The issue is
    # in pipeline.py +999 causes an error because if I use `process_frame`
    # directly, without setting the aiko.message object to something I
    # get an attribute error when .publish is called
    class MockMessage:
        def publish(self, *args, **kwargs):
            pass

        def subscribe(self, *args, **kwargs):
            pass

        def unsubscribe(self, *args, **kwargs):
            pass

    aiko_main.message = MockMessage()



from pathlib import Path

from ..core import DEFAULT_PAGE_SIZE

# These 4 imports are made availiable from this file
# for backwards compatability.
from .base_models import *
from .cropping import *
from .dataset import *
from .formats.aws_s3.writer import AwsS3Writer
from .formats.coco.common import CocoCategory, bbox_to_wkt, segmentation_to_wkt
from .formats.coco.reader import CocoReader
from .formats.coco.writer import CocoWriter
from .formats.highlighter.reader import HighlighterAssessmentsReader
from .formats.highlighter.writer import HighlighterAssessmentsWriter
from .formats.torch_image_folder.reader import TorchImageFolderReader
from .formats.torch_image_folder.writer import TorchImageFolderWriter
from .formats.yolo.writer import YoloWriter
from .splits import SUPPORTED_SPLIT_FNS, RandomSplitter, get_split_fn

READERS = {
    CocoReader.format_name: CocoReader,
    HighlighterAssessmentsReader.format_name: HighlighterAssessmentsReader,
    TorchImageFolderReader.format_name: TorchImageFolderReader,
}

WRITERS = {
    CocoWriter.format_name: CocoWriter,
    HighlighterAssessmentsWriter.format_name: HighlighterAssessmentsWriter,
    TorchImageFolderWriter.format_name: TorchImageFolderWriter,
    AwsS3Writer.format_name: AwsS3Writer,
}


def get_reader(name):
    return READERS[name]


def get_writer(name):
    return WRITERS[name]


def read_dataset_from_highlighter(
    client: "HLClient",
    dataset_id: int,
    datasets_cache_dir: Path = None,
    data_files_cache_dir: Path = None,
    page_size: int = DEFAULT_PAGE_SIZE,
):
    """Reads and initializes a dataset stored in HL"""
    ds = Dataset.read_from(
        dataset_format=DatasetFormat.HIGHLIGHTER_DATASET,
        client=client,
        dataset_id=dataset_id,
        datasets_cache_dir=datasets_cache_dir,
        data_files_cache_dir=data_files_cache_dir,
        page_size=page_size,
    )

    return ds


def write_assessments_to_highlighter(client, ds, workflow_id, user_id=None):
    """Write a dataset as assessments to a highlighter workflow

    Args:
        client (highlighter.gql_client.HLClient): graphql client
        ds (highlighter.datasets.Dataset): dataset to write
        workflow_id (int): id of workflow to write new assessments to
        user_id (int): id of the user to be the 'submitter' e.g. 493, default=None

    Returns:
        _type_: dataset with updated assessment id's and hashes (edited in place)
    """
    writer = get_writer("highlighter_assessments")(client, workflow_id, user_id=user_id)

    writer.write(ds)

    return ds



#!/usr/bin/env python3
#
# Usage
# ~~~~~
# LOG_LEVEL=DEBUG ./video_to_images.py

import aiko_services as aiko

FRAME_RATE = 0   # Process flat-out without delay
IMAGE_OUTPUT_PATHNAME = "z_output/image_{:06d}.jpg"
VIDEO_INPUT_PATHNAME = "astra_brief.mp4"

ELEMENTS_VIDEO = "aiko_services.elements.video_io"
ELEMENTS_IMAGE = "aiko_services.elements.image_io"

pipeline_definition = [
    {   "name": "VideoReadFile", "module": ELEMENTS_VIDEO,
        "successors": ["ImageOverlay"],
        "parameters": {
            "video_pathname": VIDEO_INPUT_PATHNAME
        }
    },
    {   "name": "ImageOverlay", "module": ELEMENTS_IMAGE,
        "successors": ["ImageWriteFile"],
        "parameters": {
        "colors": {
            "astra": [100, 0, 0],
            "bailey": [0, 100, 0],
            "ty": [0, 0, 100]
            },
        "text_color": "yellow"
        }
    },
    {   "name": "ImageWriteFile", "module": ELEMENTS_IMAGE,
        "parameters": {
            "image_pathname": IMAGE_OUTPUT_PATHNAME
        }
    }
]

# if __name__ == "__main__":
#   Pipeline_2020(pipeline_definition, FRAME_RATE).run()



import aiko_services as aiko

__all__ = ["HLDataSchemeHTTP"]


class HLDataSchemeHTTP(aiko.DataScheme):

    def create_sources(self, stream, data_sources, frame_generator=None, use_create_frame=True):

        urls = []
        for data_source in data_sources:
            try:
                urls.append(data_source.replace("hlhttp://", ""))

            except Exception as e:
                diagnostic = f'Error loading file, "{e}"'
                return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

        pipeline_element = self.pipeline_element
        if use_create_frame:
            """
            - Not sure how to determine the frame_data keys ahead of time.
            - Not sure how this makes sense for DataSource Elements that
              produce many frames for one file, ie VideoFileRead
            """
            NotImplementedError()
        else:
            task_id = None
            stream.variables["source_paths_generator"] = iter([(u, task_id) for u in urls])
            rate, _ = pipeline_element.get_parameter("rate", default=None)
            rate = float(rate) if rate else None
            pipeline_element.create_frames(stream, frame_generator, rate=rate)
        return aiko.StreamEvent.OKAY, {}


HLDataSchemeHTTP.add_data_scheme("hlhttp", HLDataSchemeHTTP)



import json
from datetime import datetime, timedelta
from enum import Enum
from logging import warning
from os import PathLike
from pathlib import Path
from typing import Generator, List, Optional
from urllib.parse import urlparse

from highlighter.core.logging import get_default_logger

try:
    import cv2
except ModuleNotFoundError as _:
    cv2 = None

import numpy as np
from PIL import Image

from highlighter.client import download_bytes
from highlighter.client.base_models import DataFile
from highlighter.client.io import (
    _pil_open_image_bytes,
    _pil_open_image_path,
    _pil_open_image_url,
)
from highlighter.core.exceptions import require_package

from .base_capability import DataSourceCapability

__all__ = [
    "ImageDataSource",
    "TextDataSource",
    "JsonArrayDataSource",
    "VideoDataSource",
]


def _iter_buffer(
    buffer,
    sep: bytes,
) -> Generator[bytes, None, None]:
    """
    Streams data from stdin until the Nth occurrence of a separator byte sequence.

    sep: The separator byte sequence (e.g., b'\n' or b'\x00\x01')
    """
    if sep is None or len(sep) == 0:
        raise ValueError("Separator must be a non-empty bytes object")

    byte_result = bytearray()
    buffer_remainder = bytearray()

    while True:
        chunk = buffer.read(4096)  # Adjusted for testing, increase for performance
        if not chunk:
            if byte_result:
                byte_result = bytes(byte_result)
                yield byte_result
            break

        buffer_remainder.extend(chunk)

        while True:
            sep_index = buffer_remainder.find(sep)
            if sep_index == -1:
                break

            byte_result.extend(buffer_remainder[:sep_index])
            yield bytes(byte_result)

            byte_result.clear()
            buffer_remainder = buffer_remainder[sep_index + len(sep) :]

        byte_result.extend(buffer_remainder)
        buffer_remainder.clear()


class TextFrameIterator:

    def __init__(
        self,
        source_buffers=None,
        source_urls=None,
        logger=None,
    ):

        self.logger = logger if logger is not None else get_default_logger("TextFrameIterator")
        self.source_buffers = source_buffers
        self.source_urls = source_urls

        if self.source_urls is not None:
            self.logger.info("TextFrameIterator using source_url")
            self._source = iter([str(u) for u, _ in self.source_urls])
        elif self.source_buffers is not None:

            def iter_buffers(bufs):
                for b in bufs:
                    for item in _iter_buffer(b, "===END===".encode("utf-8")):
                        yield item.decode()

            self._source = iter_buffers(self.source_buffers)
            self.logger.info(f"TextFrameIterator using source_buffer")
        else:
            raise ValueError("Must provide source_buffer &| source_url")

        self.frame_index = 0

    def _is_url(self, p):
        return all([urlparse(p), urlparse(p).netloc])

    def _is_local_path(self, p):
        return Path(p).exists()

    def _read_text(self, text_src):
        if self._is_local_path(text_src):
            with open(text_src, "r") as f:
                text = f.read()
            original_source_url = text_src
        elif self._is_url(text_src):
            text = download_bytes(text_src).decode("utf-8")
            original_source_url = text_src
        else:
            text = text_src
            original_source_url = None
        return text, original_source_url

    def __iter__(self):
        return self

    def __next__(self):
        text_src = next(self._source)
        text_content, original_source_url = self._read_text(text_src)

        data_file = DataFile(
            # file_id=self.ds.id,
            content=text_content,
            content_type="text",
            media_frame_index=0,
            original_source_url=original_source_url,
        )
        return data_file


class TextDataSource(DataSourceCapability):
    """

    TODO: Check/update this

    Example:
        # process a single string
        hl agent start --data-source TextDataSource PIPELINE.json "tell me a joke."

        # process many text files
        ToDo

        # Read from stdin
        cat file | hl agent start --data-source TextDataSource -sp read_stdin=true PIPELINE.json
    """

    stream_media_type = "text"

    class DefaultStreamParameters(DataSourceCapability.DefaultStreamParameters):
        byte_encoding: Optional[str] = "utf-8"

    @property
    def byte_encoding(self) -> str:
        value, _ = self._get_parameter("byte_encoding")
        return value

    def get_text_frame_generator(self, stream):

        if "source_paths_generator" in stream.variables:
            for frame_data in TextFrameIterator(
                source_buffers=None,
                source_urls=stream.variables["source_paths_generator"],
                logger=self.logger,
            ):
                yield frame_data
        elif "source_buffers" in stream.variables:
            for frame_data in TextFrameIterator(
                source_buffers=stream.variables["source_buffers"],
                source_urls=None,
                logger=self.logger,
            ):
                yield frame_data
        elif "records" in stream.variables:
            raise NotImplementedError("#### ToDo records ####")
        else:
            raise NotImplementedError("#### ToDo ####")

    def get_next_frame_data(self, stream):

        frame_generator = stream.variables.get("text_frame_generator", None)
        if frame_generator is None:
            frame_generator = self.get_text_frame_generator(stream)
            stream.variables["text_frame_generator"] = frame_generator

        data_file = next(frame_generator)
        entities = {}  # ToDo, how/where to load existing entities from
        return data_file, entities


class JsonArrayFrameIterator:

    def __init__(self, source_buffers=None, source_urls=None, key="", logger=None):

        self.logger = logger if logger is not None else get_default_logger("JsonArrayFrameIterator")
        self.source_urls = source_urls
        self.source_buffers = source_buffers
        self.key = key

        if self.source_urls is not None:
            self.logger.info("Using source_urls")
            self._source = self._iter_source_urls(self.source_urls)

        elif self.source_buffers is not None:
            self.logger.info("Using source_buffers")
            self._source = self._iter_source_buffers(self.source_buffers)
        else:
            raise ValueError("Must provide source_buffer &| source_url")

    def _is_url(self, p):
        return all([urlparse(p), urlparse(p).netloc])

    def _is_local_path(self, p):
        return Path(p).exists()

    def _iter_source_urls(self, source_urls):
        frame_counter = 0
        for url in source_urls:
            url = str(url)
            if self._is_url(url):
                arr = json.loads(download_bytes(url).decode("utf-8"))
            elif self._is_local_path(url):
                with open(url, "r") as f:
                    arr = json.load(f)
            else:
                raise ValueError(f"Unable to load {url} as json")

            for k in self.key.split("."):
                arr = arr[k]

            for item in arr:
                yield item, url, frame_counter
                frame_counter += 1

    def _iter_source_buffers(self, source_buffers):
        frame_counter = 0
        for buf in source_buffers:
            for item in _iter_buffer(buf, "===END===".encode("utf-8")):
                arr = json.loads(item)

                for k in self.key.split("."):
                    arr = arr[k]

                for item in arr:
                    yield item, None, frame_counter
                    frame_counter += 1

    def __iter__(self):
        return self

    def __next__(self):

        content, original_source_url, frame_index = next(self._source)
        data_file = DataFile(
            # file_id=self.ds.id,
            content=content,
            content_type="text",
            media_frame_index=frame_index,
            original_source_url=original_source_url,
        )
        return data_file


class JsonArrayDataSource(DataSourceCapability):

    stream_media_type = "text"

    class DefaultStreamParameters(DataSourceCapability.DefaultStreamParameters):
        key: str = ""

    @property
    def key(self) -> str:
        value, _ = self._get_parameter("key")
        return value

    def get_json_array_frame_generator(self, stream):

        if "source_paths_generator" in stream.variables:
            source_urls = [path for path, _ in stream.variables["source_paths_generator"]]
            for frame_data in JsonArrayFrameIterator(
                source_buffers=None,
                source_urls=source_urls,
                key=self.key,
                logger=self.logger,
            ):
                yield frame_data
        elif "source_buffers" in stream.variables:
            for frame_data in JsonArrayFrameIterator(
                source_buffers=stream.variables["source_buffers"],
                source_urls=None,
                key=self.key,
                logger=self.logger,
            ):
                yield frame_data
        elif "records" in stream.variables:
            raise NotImplementedError("#### ToDo records ####")
        else:
            raise NotImplementedError("#### ToDo ####")

    def get_next_frame_data(self, stream):

        json_array_frame_generator = stream.variables.get("json_array_frame_generator", None)
        if json_array_frame_generator is None:
            json_array_frame_generator = self.get_json_array_frame_generator(stream)
            stream.variables["json_array_frame_generator"] = json_array_frame_generator

        data_file = next(json_array_frame_generator)
        entities = {}  # ToDo, how/where to load existing entities from
        return data_file, entities


class OutputType(str, Enum):
    numpy = "numpy"
    pillow = "pillow"


class ImageFrameIterator:
    def __init__(
        self,
        source_buffers=None,
        source_urls=None,
        output_type: OutputType = OutputType.numpy,
        logger=None,
    ):

        self.logger = logger if logger is not None else get_default_logger("ImageFrameIterator")
        self.source_buffers = source_buffers
        self.source_urls = source_urls
        self.output_type = output_type

        if self.source_urls is not None:
            self.logger.info("ImageFrameIterator using source_url")
            self._source = iter([str(u) for u, _ in self.source_urls])
        elif self.source_buffers is not None:

            def iter_buffers(bufs):
                for b in bufs:
                    for item in _iter_buffer(b, "===END===".encode("utf-8")):
                        yield item

            self._source = iter_buffers(self.source_buffers)
            self.logger.info(f"ImageFrameIterator using source_buffer")
        else:
            raise ValueError("Must provide source_buffer &| source_url")

        self.frame_index = 0

    def _is_url(self, p):
        return all([urlparse(p), urlparse(p).netloc])

    def _is_local_path(self, p):
        return Path(p).exists()

    def __iter__(self):
        return self

    def _read_image(self, img_src):
        if isinstance(img_src, str) and self._is_local_path(img_src):
            img = _pil_open_image_path(img_src)
            original_source_url = img_src
        elif isinstance(img_src, str) and self._is_url(img_src):
            img = _pil_open_image_url(img_src)
            original_source_url = img_src
        else:
            img = _pil_open_image_bytes(img_src)
            original_source_url = None
        return img, original_source_url

    def __next__(self):
        img_src = next(self._source)
        img, original_source_url = self._read_image(img_src)

        if self.output_type == OutputType.numpy:
            img = np.array(img, dtype=np.uint8)

        data_file = DataFile(
            # file_id=self.ds.id,
            content=img,
            content_type="image",
            media_frame_index=0,
            original_source_url=original_source_url,
        )
        return data_file


class ImageDataSource(DataSourceCapability):
    """

    Example:
        # process a single image
        hl agent start PIPELINE.json image.jpg

        # process many images
        find image/dir/ -n "*.jpg" | hl agent start PIPELINE.json
    """

    stream_media_type = "image"

    class DefaultStreamParameters(DataSourceCapability.DefaultStreamParameters):
        output_type: OutputType = OutputType.numpy

    @property
    def output_type(self) -> OutputType:
        value, _ = self._get_parameter("output_type")
        return value

    def get_image_frame_generator(self, stream):

        if "source_paths_generator" in stream.variables:
            for frame_data in ImageFrameIterator(
                source_buffers=None,
                source_urls=stream.variables["source_paths_generator"],
                output_type=self.output_type,
                logger=self.logger,
            ):
                yield frame_data
        elif "source_buffers" in stream.variables:
            for frame_data in ImageFrameIterator(
                source_buffers=stream.variables["source_buffers"],
                source_urls=None,
                output_type=self.output_type,
                logger=self.logger,
            ):
                yield frame_data
        elif "records" in stream.variables:
            raise NotImplementedError("#### ToDo records ####")
        else:
            raise NotImplementedError("#### ToDo ####")

    def get_next_frame_data(self, stream):

        frame_generator = stream.variables.get("image_frame_generator", None)
        if frame_generator is None:
            frame_generator = self.get_image_frame_generator(stream)
            stream.variables["image_frame_generator"] = frame_generator

        data_file = next(frame_generator)
        entities = {}  # ToDo, how/where to load existing entities from
        return data_file, entities


@require_package(cv2, "cv2", "opencv")
class VideoFrameIterator:

    def __init__(
        self, source_buffers=None, source_urls=None, output_type: OutputType = OutputType.numpy, logger=None
    ):

        self.logger = logger if logger is not None else get_default_logger("VideoFrameIterator")
        self.output_type = output_type
        self.source_urls = source_urls
        self.source_buffers = source_buffers
        self.start_time = datetime.now()
        self.frame_index = 0

        if self.source_urls is not None:
            self.logger.info("Using VideoCapture")
            self._source = iter([str(u) for u in self.source_urls])
            self._cap_cls = cv2.VideoCapture

        elif self.source_buffers is not None:
            from highlighter.agent.capabilities._buffered_video_reader import (
                BufferedVideoCapture,
            )

            self.logger.info("Using BufferedVideoCapture")
            self._source = iter(self.source_buffers)
            self._cap_cls = BufferedVideoCapture
        else:
            raise ValueError("Must provide source_buffer &| source_url")

        self._cap, self._original_source_url = self._init_video_capture()

    def _init_video_capture(self):
        src = next(self._source)
        _cap = self._cap_cls(src)
        if not _cap.isOpened():
            raise ValueError("Cannot open video file or buffer")
        if isinstance(src, str):
            _original_source_url = src
        else:
            _original_source_url = None
        return _cap, _original_source_url

    def __iter__(self):
        return self

    def __next__(self):

        ret, frame_img = self._cap.read()
        if not ret:  # No more frames to read
            self._cap.release()
            self._cap, self._original_source_url = self._init_video_capture()
            ret, frame_img = self._cap.read()
            if not ret:
                self._cap.release()
                raise StopIteration

        frame_img = cv2.cvtColor(frame_img, cv2.COLOR_BGR2RGB)

        if self.output_type == OutputType.pillow:
            frame_img = Image.fromarray(frame_img)

        timestamp = self._cap.get(cv2.CAP_PROP_POS_MSEC)
        data_file = DataFile(
            # file_id=self.ds.id,
            content=frame_img,
            content_type="image",
            recorded_at=self.start_time + timedelta(milliseconds=timestamp),
            media_frame_index=self.frame_index,
            original_source_url=self._original_source_url,
        )
        self.frame_index += 1
        return data_file

    def __del__(self):
        if self._cap.isOpened():
            self._cap.release()

    def get_total_frames(self):
        return int(self._cap.get(cv2.CAP_PROP_FRAME_COUNT))

    def extract_frames_by_id(
        self, frame_ids: List[int], frame_save_dir: PathLike, filename: str
    ) -> List[int]:

        frame_ids = sorted(frame_ids, reverse=True)

        total_frames = self.get_total_frames()
        invalid_ids = [fid for fid in frame_ids if fid < 0 or fid >= total_frames]
        if invalid_ids:
            warning(f"These frame IDs are invalid (out of range 0-{total_frames-1}): {invalid_ids}")
            frame_ids = [fid for fid in frame_ids if fid >= 0 and fid < total_frames]

        extracted_frame_ids = []
        frame_id = frame_ids.pop()
        _frame_save_dir = Path(frame_save_dir)
        _frame_save_dir.mkdir(parents=True, exist_ok=True)
        for frame in self:
            if frame.media_frame_index == frame_id:
                save_path = _frame_save_dir / filename.format(frame_id)
                cv2.imwrite(str(save_path), frame.content)
                extracted_frame_ids.append(frame_id)
                if frame_ids:
                    frame_id = frame_ids.pop()
                else:
                    break
        return extracted_frame_ids


class VideoDataSource(DataSourceCapability):

    stream_media_type = "video"

    class DefaultStreamParameters(DataSourceCapability.DefaultStreamParameters):
        output_type: OutputType = OutputType.numpy

    @property
    def output_type(self) -> OutputType:
        value, _ = self._get_parameter("output_type")
        return value

    def get_video_frame_generator(self, stream):

        if "source_paths_generator" in stream.variables:
            source_urls = [path for path, task_id in stream.variables["source_paths_generator"]]
            for frame_data in VideoFrameIterator(
                source_buffers=None,
                source_urls=source_urls,
                output_type=self.output_type,
                logger=self.logger,
            ):
                yield frame_data
        elif "source_buffers" in stream.variables:
            for frame_data in VideoFrameIterator(
                source_buffers=stream.variables["source_buffers"],
                source_urls=None,
                output_type=self.output_type,
                logger=self.logger,
            ):
                yield frame_data
        elif "records" in stream.variables:
            raise NotImplementedError("#### ToDo records ####")
        else:
            raise NotImplementedError("#### ToDo ####")

    def get_next_frame_data(self, stream):

        video_frame_generator = stream.variables.get("video_frame_generator", None)
        if video_frame_generator is None:
            video_frame_generator = self.get_video_frame_generator(stream)
            stream.variables["video_frame_generator"] = video_frame_generator

        data_file = next(video_frame_generator)
        entities = {}  # ToDo, how/where to load existing entities from
        return data_file, entities



# Usage: File
# ~~~~~~~~~~~
# aiko_pipeline create pipelines/video_pipeline_0.json -s 1 -sr -ll debug
#
# aiko_pipeline create pipelines/video_pipeline_0.json -s 1 -p rate 4.0
#
# aiko_pipeline create pipelines/video_pipeline_0.json -s 1  \
#   -p VideoReadFile.data_batch_size 8
#
# aiko_pipeline create pipelines/video_pipeline_0.json -s 1  \
#   -p VideoReadFile.data_sources file://data_in/in_{}.mp4
#
# aiko_pipeline create pipelines/video_pipeline_0.json -s 1  \
#   -p VideoWriteFile.path "file://data_out/out_{:02d}.mp4"
#
# aiko_pipeline create pipelines/video_pipeline_0.json -s 1  \
#   -p VideoReadFile.data_sources file://data_in/in_00.mp4   \
#   -p ImageResize.resolution 320x240                        \
#   -p VideoWriteFile.data_targets file://data_out/out_00.mp4
#
# Drop frame test
# ~~~~~~~~~~~~~~~
# aiko_pipeline create pipelines/video_pipeline_1.json -s 1 -ll debug
#
# To Do
# ~~~~~
# - Implement "VideoReadFile.data_batch_size" in "frame_generator()"
#
# - Refactor optional module import into common function (see image_io.py)
#
# - Implement start_frame (first) and stop_frame (last) parameters
# - Improve PipelineElement VideoSample()
#   - Rather than sample by frame ... sample by image count in [images] input
#   - Provide video_sample() for use by VideoFileRead and VideoSample
#
# - Implement ...
#     video_capture = cv2.VideoCapture(path)
#     width = int(video_capture.get(cv2.CAP_PROP_FRAME_WIDTH))
#     height = int(video_capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
#     length = int(video_capture.get(cv2.CAP_PROP_FRAME_COUNT))
#     frame_rate = int(video_capture.get(cv2.CAP_PROP_FPS))
#
# - VideoReadFile should accept a DataSource type ...
#   - URL: "file://" and media_type: "mp4", etc
#
# - Metrics: Determine what to metrics to capture, e.g frame rates ?
#
# - Design video windowing, i.e collecting multiple frame for ML processing
#   together, e.g gesture analysis
#
# - "Batching" images for CPU-GPU memory transfer efficiently for nVidia
#   - For both images and video ... depending upon ML model
#
# - Consider VideoShow() GUI ...
#   - Why aren't the cv2.imshow() graphical icons appearing about the image ?
#   - cv2.createTrackbar() ?
#   - Integrate with tkinter ?

from typing import Tuple
from pathlib import Path

import aiko_services as aiko

__all__ = [
    "VideoOutput", "VideoReadFile", "VideoSample", "VideoShow", "VideoWriteFile"
]

_LOGGER = aiko.get_logger(__name__)

_CV2_IMPORTED = False
try:
    import cv2
    _CV2_IMPORTED = True
except ModuleNotFoundError:  # TODO: Optional warning flag
    diagnostic = "video_io.py: Couldn't import cv2 module"
#   print(f"WARNING: {diagnostic}")
#   _LOGGER.warning(diagnostic)
#   raise ModuleNotFoundError(
#       'opencv-python package not installed.  '
#       'Install aiko_services with --extras "opencv" '
#       'or install opencv-python manually to use the "video_io" module')

_NUMPY_IMPORTED = False
try:
    import numpy as np
    _NUMPY_IMPORTED = True
except ModuleNotFoundError:  # TODO: Optional warning flag
    diagnostic = "video_io.py: Couldn't import numpy module"
#   print(f"WARNING: {diagnostic}")
#   _LOGGER.warning(diagnostic)

# --------------------------------------------------------------------------- #
# Useful for Pipeline output that should be all of the images processed

class VideoOutput(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("video_output:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, images) -> Tuple[aiko.StreamEvent, dict]:
        return aiko.StreamEvent.OKAY, {"images": images}

# --------------------------------------------------------------------------- #
# VideoReadFile is a DataSource which supports ...
# - Individual video files
# - Directory of video files with an optional filename filter
# - TODO: Archive (tgz, zip) of video files with an optional filename filter
#
# parameter: "data_sources" is the read file path, format variable: "video_id"
#
# Note: Only supports Streams with "data_sources" parameter

class VideoReadFile(aiko.DataSource):  # PipelineElement
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("video_read_file:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def start_stream(self, stream, stream_id):
        stream.variables["video_capture"] = None
        stream.variables["video_frame_generator"] = None
        return super().start_stream(stream, stream_id,
            frame_generator=self.frame_generator, use_create_frame=False)

    def video_frame_iterator(self, video_capture):
        while True:
            status, image_bgr = video_capture.read()
            if not status:
                break
            image_rgb = cv2.cvtColor(image_bgr, cv2.COLOR_BGR2RGB)
            yield image_rgb

    def frame_generator(self, stream, frame_id):
        video_frame_generator = stream.variables["video_frame_generator"]
        while True:
            if not video_frame_generator:
                try:
                    path, file_id = next(
                        stream.variables["source_paths_generator"])
                except StopIteration:
                    stream.variables["source_paths_generator"] = None
                    diagnostic = "End of video file(s)"
                    return aiko.StreamEvent.STOP, {"diagnostic": diagnostic}

                if not path.is_file():
                    diagnostic = f'path "{path}" must be a file'
                    return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

            # TODO: handle integer video "path" ... maybe ?
            #   if isinstance(path, str) and path.isdigit():
            #       path = int(str(path))

                video_capture = cv2.VideoCapture(str(path))
                if not video_capture.isOpened():
                    diagnostic = f"Couldn't open video file: {path}"
                    return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}
                stream.variables["video_capture"] = video_capture
                video_frame_generator = self.video_frame_iterator(video_capture)
                stream.variables["video_frame_generator"] =  \
                    video_frame_generator

            try:
                image_rgb = next(video_frame_generator)
                return aiko.StreamEvent.OKAY, {"images": [image_rgb]}
            except StopIteration:
                video_frame_generator = None
                stream.variables["video_frame_generator"] = None

    def process_frame(self, stream, images) -> Tuple[aiko.StreamEvent, dict]:
        self.logger.debug(f"{self.my_id()}")
        return aiko.StreamEvent.OKAY, {"images": images}

    def stop_stream(self, stream, stream_id):
        video_capture = stream.variables["video_capture"]
        if video_capture and video_capture.isOpened():
            video_capture.release()
            stream.variables["video_capture"] = None
        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #

class VideoSample(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("video_sample:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, images) -> Tuple[aiko.StreamEvent, dict]:
        sample_rate, _ = self.get_parameter("sample_rate", 1)
        if stream.frame_id % sample_rate:
            self.logger.debug(f"{self.my_id()}: frame dropped")
            return aiko.StreamEvent.DROP_FRAME, {}
        else:
            self.logger.debug(f"{self.my_id()}: frame not dropped")
            return aiko.StreamEvent.OKAY, {"images": images}

# --------------------------------------------------------------------------- #
# TODO: Change color, title and resolution

class VideoShow(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("video_show:0")
        context.get_implementation("PipelineElement").__init__(self, context)

#   def slider_handler(self, value):
#       print(f"slider: {value}")
#       cv2.imshow(title, image_bgr)  # TODO: "title" ?!?

    def process_frame(self, stream, images) -> Tuple[aiko.StreamEvent, dict]:
        for image in images:
            if not isinstance(image, np.ndarray):
                image = np.array(image)  # RGB

            grayscale = len(image.shape) == 2
            if grayscale:
                image_bgr = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
            else:
                image_bgr = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)

            title, _ = self.get_parameter("title", "Video")
            cv2.namedWindow(title)
        #   cv2.createTrackbar("Slider", title, 0, 9, self.slider_handler)
            cv2.imshow(title, image_bgr)
            if stream.frame_id == 0:
                position, _ = self.get_parameter("position", "1280:0")
                position_x, position_y = position.split(":")
                cv2.moveWindow(title, int(position_x), int(position_y))
            if cv2.waitKey(1) & 0xFF == ord("x"):
                system_exit, _ = self.get_parameter("system_exit", False)
                diagnostic = "VideoShow exit"
                if system_exit:
                    raise SystemExit(diagnostic)
                else:
                    return aiko.StreamEvent.STOP, {"diagnostic": diagnostic}
        return aiko.StreamEvent.OKAY, {}

    def stream_stop_handler(self, stream, stream_id):
        cv2.destroyAllWindows()  # TODO: when all Streams stopped
        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #
# VideoWriteFile is a DataTarget that writes images to a video file
#
# parameter: "data_targets" is the write file path, format variable: "frame_id"
#
# Note: Only supports Streams with "data_targets" parameters

class VideoWriteFile(aiko.DataTarget):  # PipelineElement
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("video_write_file:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def start_stream(self, stream, stream_id):
        stream_event, diagnostic = super().start_stream(stream, stream_id)

        if stream_event == aiko.StreamEvent.OKAY:
            path = stream.variables["target_path"]
            if stream.variables["target_path_template"]:
                path = path.format(stream.variables["target_file_id"])
            self.logger.debug(f"{self.my_id()}: {path}")
            stream.variables["video_path"] = path
            stream.variables["video_writer"] = None
            stream_event = aiko.StreamEvent.OKAY
            diagnostic = {}
        return stream_event, diagnostic

# Show codec list by invoking cv2.VideoWriter(..., fourcc=-1)
# video_output.avi: format: XVID
# video_output.mp4: format: MP4V, DIVX, H264, X264
# Linux:   DIVX, XVID, MJPG, X264, WMV1, WMV2
# WIndows: DIVX

    def _create_video_writer(
        self, path, resolution, format="MP4V", frame_rate=30.0):

        format, _ = self.get_parameter("format", format)
        format = cv2.VideoWriter_fourcc(*format)
        frame_rate, _ = self.get_parameter("frame_rate", frame_rate)
        resolution, _ = self.get_parameter("resolution", resolution)
        if isinstance(resolution, str):
            width, height = resolution.split("x")
            resolution = (int(width), int(height))
        parent_path = Path(path).parent
        parent_path.mkdir(exist_ok=True, parents=True)
        path = str(path)
        video_writer = cv2.VideoWriter(
            path, format, float(frame_rate), resolution)
        return video_writer

    def process_frame(self, stream, images) -> Tuple[aiko.StreamEvent, dict]:
        self.logger.debug(f"{self.my_id()}")

        if stream.variables["video_writer"]:
            video_writer = stream.variables["video_writer"]
        else:
            path = stream.variables["video_path"]
            resolution = (images[0].shape[1], images[0].shape[0])
            video_writer = self._create_video_writer(path, resolution)
            stream.variables["video_writer"] = video_writer

        for image in images:
            image_bgr = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
            video_writer.write(image_bgr)

        return aiko.StreamEvent.OKAY, {}

    def stop_stream(self, stream, stream_id):
        if stream.variables["video_writer"]:
            stream.variables["video_writer"].release()
            stream.variables["video_writer"] = None
        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #
